import pytest

import numpy as np

from pyquil.quil import Program
from pyquil.quilatom import (
    MemoryReference,
    Parameter,
    FormalArgument,
    Frame,
)
from pyquil.quiltwaveforms import (
    FlatWaveform,
    GaussianWaveform,
    DragGaussianWaveform,
    ErfSquareWaveform,
    HrmGaussianWaveform,
    BoxcarAveragerKernel,
)
from pyquil.quiltcalibrations import (
    CalibrationError,
    CalibrationMatch,
    fill_placeholders,
    match_calibration,
)
from pyquil.quilbase import (
    Gate,
    DefCalibration,
    ShiftPhase,
    DelayQubits,
    Qubit,
)


def test_waveform_samples():
    # this is a very naive check: can we sample from the built-in template
    # waveforms?
    duration = 1e-6
    waveforms = [
        FlatWaveform(duration=duration, iq=1.0),
        FlatWaveform(duration=duration, iq=1.0 + 2.0j),
        GaussianWaveform(duration=duration, fwhm=2.0, t0=1.0),
        DragGaussianWaveform(duration=duration, fwhm=duration / 4, t0=duration / 2, anh=5.0, alpha=3.0),
        HrmGaussianWaveform(
            duration=duration,
            fwhm=duration / 4,
            t0=duration / 2,
            anh=5.0,
            alpha=3.0,
            second_order_hrm_coeff=0.5,
        ),
        ErfSquareWaveform(duration=duration, risetime=duration / 8, pad_left=0.0, pad_right=0.0),
        BoxcarAveragerKernel(duration=duration),
    ]

    rates = [int(1e9), 1e9, 1e9 + 0.5]

    for rate in rates:
        for wf in waveforms:
            assert wf.samples(rate) is not None


def test_waveform_samples_optional_args():
    def flat(**kwargs):
        return FlatWaveform(duration=1e-8, iq=1.0, **kwargs).samples(1e9)

    assert np.array_equal(2.0 * flat(), flat(scale=2.0))
    assert np.array_equal(np.exp(1j) * flat(), flat(phase=1.0))


def _match(cal, instr):
    # get the first line, remove trailing colon if present
    cal_header = cal.splitlines()[0].strip().replace(":", "")
    # convert to quil ast
    full_calibration = cal_header + ":\n    NOP\n\n"
    cal = Program(full_calibration).calibrations[0]

    # we pull the last instr (omitting implicit DECLARE)
    instr = Program(instr)[-1]
    return match_calibration(instr, cal)


def test_simple_gate_calibration_match():
    matches = [
        ("DEFCAL X 0", "X 0"),
        ("DEFCAL X q", "X 0"),
        ("DEFCAL FOO q 0", "FOO 1 0"),
        ("DEFCAL FOO q 0", "FOO 0 0"),
    ]

    for cal, instr in matches:
        assert _match(cal, instr) is not None

    assert _match(cal, instr).settings[FormalArgument("q")] == Qubit(0)

    mismatches = [
        ("DEFCAL X 0", "X 1"),
        ("DEFCAL Y 0", "X 0"),
        # we are case sensitive
        ("DEFCAL foo 0", "FOO 0"),
    ]

    for cal, instr in mismatches:
        assert _match(cal, instr) is None


def test_parametric_calibration_match():
    matches = [
        ("DEFCAL RX(0.0) 0", "RX(0.0) 0"),
        ("DEFCAL RX(0.0) 0", "RX(0*pi) 0"),
        ("DEFCAL RX(pi/2) 0", "RX(pi/2) 0"),
        ("DEFCAL RX(pi/2) q", "RX(pi/2) 0"),
        ("DEFCAL RX(pi/2) q", "RX(pi/2) 1"),
        ("DEFCAL RZ(pi/2) 0", "RZ(pi/2) 0"),
        ("DEFCAL RX(%theta) 0", "RX(pi) 0"),
    ]

    for cal, instr in matches:
        assert _match(cal, instr) is not None

    assert np.isclose(_match(cal, instr).settings[Parameter("theta")], np.pi)

    mismatches = [
        ("DEFCAL RX(pi) q", "RX(0) 0"),
        ("DEFCAL RX(pi) 0", "RX(0) 0"),
        ("DEFCAL RX(pi) 0", "RY(pi) 0"),
    ]

    for cal, instr in mismatches:
        assert not _match(cal, instr)


def test_memory_reference_parameter():
    assert _match("DEFCAL RX(%theta) q", "DECLARE theta REAL\nRZ(theta) 0") is None


def test_measure_calibration_match():
    matches = [
        ("DEFCAL MEASURE 0", "MEASURE 0"),
        ("DEFCAL MEASURE q", "MEASURE 0"),
        ("DEFCAL MEASURE 0 b", "DECLARE ro BIT\nMEASURE 0 ro[0]"),
        ("DEFCAL MEASURE q b", "DECLARE ro BIT\nMEASURE 1 ro[0]"),
    ]

    for cal, instr in matches:
        assert _match(cal, instr) is not None

    assert _match(cal, instr).settings[FormalArgument("q")] == Qubit(1)
    assert _match(cal, instr).settings[FormalArgument("b")] == MemoryReference("ro")

    mismatches = [
        ("DEFCAL MEASURE 1", "MEASURE 0"),
        ("DEFCAL MEASURE 0 b", "MEASURE 0"),
        ("DEFCAL MEASURE q", "DECLARE ro BIT\nMEASURE 0 ro[0]"),
    ]

    for cal, instr in mismatches:
        assert _match(cal, instr) is None


def test_apply_match_shift_phase():
    settings = {FormalArgument("q"): Qubit(0), Parameter("theta"): np.pi}

    instr = ShiftPhase(Frame([FormalArgument("q")], "ff"), Parameter("theta") / (2.0 * np.pi))

    actual = fill_placeholders(instr, settings)

    expected = ShiftPhase(Frame([Qubit(0)], "ff"), 0.5)

    assert actual == expected


def test_apply_match_delay_qubits():
    settings = {FormalArgument("q"): Qubit(0), Parameter("foo"): 1.0}

    instr = DelayQubits([Qubit(1), FormalArgument("q")], duration=Parameter("foo"))

    actual = fill_placeholders(instr, settings)

    expected = DelayQubits([Qubit(1), Qubit(0)], 1.0)

    assert actual == expected


def test_program_match_last():
    first = DefCalibration("X", [], [Qubit(0)], ["foo"])
    second = DefCalibration("X", [], [Qubit(0)], ["bar"])
    prog = Program(first, second)
    match = prog.match_calibrations(Gate("X", [], [Qubit(0)]))
    assert match == CalibrationMatch(cal=second, settings={})


@pytest.mark.parametrize(
    "program_input,gate,program_output",
    [
        (
            Program(
                """
DEFCAL RZ(%theta) q:
    SHIFT-PHASE q "rf" -%theta
"""
            ),
            Gate("RZ", [np.pi], [Qubit(0)]),
            Program('SHIFT-PHASE 0 "rf" -pi'),
        ),
        (
            Program(
                """
DEFCAL A(%theta) q:
    SHIFT-PHASE q "rf" -%theta

DEFCAL RZ(%theta) q:
    SHIFT-PHASE q "rf" -%theta
    A(%theta) q
"""
            ),
            Gate("RZ", [np.pi], [Qubit(0)]),
            Program('SHIFT-PHASE 0 "rf" -pi', 'SHIFT-PHASE 0 "rf" -pi'),
        ),
        (
            Program(
                """
DEFCAL A(%theta) q:
    RX(%theta) q

DEFCAL B(%theta) q:
    RY(%theta) q
DEFCAL RZ(%theta) q:
    A(%theta) q
    B(%theta) q
    A(%theta) q
"""
            ),
            Gate("RZ", [np.pi], [Qubit(0)]),
            Program("RX(pi) 0", "RY(pi) 0", "RX(pi) 0"),
        ),
        (
            Program(
                """
DEFCAL RX(%theta) q:
    RY(%theta) q

DEFCAL RZ(%theta) q:
    RX(%theta) q
"""
            ),
            Gate("RZ", [np.pi], [Qubit(0)]),
            Program("RY(pi) 0"),
        ),
        (
            Program(
                """
DEFCAL RY(pi) 0:
    I 0
DEFCAL RZ(0) 0:
    RY(pi) 0
DEFCAL RX(0) 0:
    RY(pi) 0
    RZ(0) 0
"""
            ),
            Gate("RX", [0], [Qubit(0)]),
            Program("I 0", "I 0"),
        ),
        (
            Program(
                """
DEFCAL RX(%theta) 0:
    RX(%theta / 2) 0

DEFCAL RX(pi / 4) 0:
    I 0
"""
            ),
            Gate("RX", [np.pi], [Qubit(0)]),
            Program("I 0"),
        ),
    ],
)
def test_program_calibrate(program_input, gate, program_output):
    calibrated = program_input.calibrate(gate)
    assert Program(calibrated) == program_output


@pytest.mark.parametrize(
    "program_text",
    (
        """
DEFCAL RZ(%theta) q:
    RZ(%theta) q
""",
        """
DEFCAL RX(%theta) q:
    RZ(%theta) q

DEFCAL RZ(%theta) q:
    RX(%theta) q
""",
        """
DEFCAL RX(%theta) q:
    RZ(0) q

DEFCAL RZ(%theta) q:
    RX(%theta) q
""",
        """
DEFCAL RX(%theta) q:
    RZ(%theta) q

DEFCAL RZ(%theta) q:
    RX(0) q
""",
        """
DEFCAL RZ(%theta) q:
    RZ(0) q

DEFCAL RZ(0) q:
    RZ(pi) q
""",
    ),
)
def test_program_calibrate_cyclic_error(program_text):
    prog = Program(program_text)
    with pytest.raises(CalibrationError):
        prog.calibrate(Gate("RZ", [np.pi], [Qubit(0)]))


def test_merge_programs_with_quilt_features():
    prog_1 = Program(
        """
DEFCAL RZ(%theta) q:
    SHIFT-PHASE q "rf" -%theta
"""
    )
    prog_2 = Program() + prog_1
    assert len(prog_2.calibrations) == 1
    prog_2 = prog_1 + Program()
    assert len(prog_2.calibrations) == 1
    prog_2 = Program()
    prog_2 += prog_1
    assert len(prog_2.calibrations) == 1

    prog_1 = Program(
        """
DEFFRAME 0 "rf":
    SAMPLE-RATE: 2.0
"""
    )
    prog_2 = Program() + prog_1
    assert len(prog_2.frames) == 1
    prog_2 = prog_1 + Program()
    assert len(prog_2.frames) == 1
    prog_2 = Program()
    prog_2 += prog_1
    assert len(prog_2.frames) == 1

    prog_1 = Program(
        """
DEFWAVEFORM foo:
    1.0, 1.0, 1.0
"""
    )
    prog_2 = Program() + prog_1
    assert len(prog_2.waveforms) == 1
    prog_2 = prog_1 + Program()
    assert len(prog_2.waveforms) == 1
    prog_2 = Program()
    prog_2 += prog_1
    assert len(prog_2.waveforms) == 1



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import ast
import sys
import functools
import inspect

from pyquil import gates
from pyquil.quil import Program
from pyquil.quilatom import MemoryReference

if sys.version_info < (3, 7):
    from pyquil.external.contextvars import ContextVar
else:
    from contextvars import ContextVar

_program_context = ContextVar("program")


def program_context() -> Program:
    """
    Returns the current program in context. Will only work from inside calls to @magicquil.
    If called from outside @magicquil will throw a ValueError

    :return: program if one exists
    """
    return _program_context.get()


def I(qubit) -> None:
    program_context().inst(gates.I(qubit))


def X(qubit) -> None:
    program_context().inst(gates.X(qubit))


def H(qubit) -> None:
    program_context().inst(gates.H(qubit))


def CNOT(qubit1, qubit2) -> None:
    program_context().inst(gates.CNOT(qubit1, qubit2))


def MEASURE(qubit) -> MemoryReference:
    program_context().inst(gates.MEASURE(qubit, MemoryReference("ro", qubit)))
    return MemoryReference("ro", qubit)


def _if_statement(test, if_function, else_function) -> None:
    """
    Evaluate an if statement within a @magicquil block.

    If the test value is a Quil MemoryReference then unwind it into quil code equivalent to an if
    then statement using jumps. Both sides of the if statement need to be evaluated and placed into
    separate Programs, which is why we create new program contexts for their evaluation.

    If the test value is not a Quil MemoryReference then fall back to what Python would normally do
    with an if statement.

    Params are:
        if <test>:
            <if_function>
        else:
            <else_function>

    NB: This function must be named exactly _if_statement and be in scope for the ast transformer
    """
    if isinstance(test, MemoryReference):
        token = _program_context.set(Program())
        if_function()
        if_program = _program_context.get()
        _program_context.reset(token)

        if else_function:
            token = _program_context.set(Program())
            else_function()
            else_program = _program_context.get()
            _program_context.reset(token)
        else:
            else_program = None

        program = _program_context.get()
        program.if_then(test, if_program, else_program)
    else:
        if test:
            if_function()
        elif else_function:
            else_function()


_EMPTY_ARGUMENTS = ast.arguments(
    args=[], posonlyargs=[], vararg=None, kwonlyargs=[], kwarg=None, defaults=[], kw_defaults=[]
)


class _IfTransformer(ast.NodeTransformer):
    """
    Transformer that unwraps the if and else branches into separate inner functions and then wraps
    them in a call to _if_statement. For example:

    .. code-block:: python

        if 1 + 1 == 2:
            print('math works')
        else:
            print('something is broken')

    would be transformed into:

    .. code-block:: python

        def _if_branch():
            print('math works')
        def _else_branch():
            print('something is broken')
        _if_statement(1 + 1 == 2, _if_branch, _else_branch)
    """

    def visit_If(self, node):
        # Must recursively visit the body of both the if and else bodies to handle any nested if
        # and else statements. This also conveniently handles elif since those are just treated
        # as a nested if/else within the else branch.
        # See: https://greentreesnakes.readthedocs.io/en/latest/nodes.html#If
        node = self.generic_visit(node)

        if_function = ast.FunctionDef(name="_if_branch", body=node.body, decorator_list=[], args=_EMPTY_ARGUMENTS)
        else_function = ast.FunctionDef(name="_else_branch", body=node.orelse, decorator_list=[], args=_EMPTY_ARGUMENTS)

        if_function_name = ast.Name(id="_if_branch", ctx=ast.Load())
        else_function_name = ast.Name(id="_else_branch", ctx=ast.Load())

        if node.orelse:
            return [
                if_function,
                else_function,
                ast.Expr(
                    ast.Call(
                        func=ast.Name(id="_if_statement", ctx=ast.Load()),
                        args=[node.test, if_function_name, else_function_name],
                        keywords=[],
                    )
                ),
            ]
        else:
            return [
                if_function,
                ast.Expr(
                    ast.Call(
                        func=ast.Name(id="_if_statement", ctx=ast.Load()),
                        args=[node.test, if_function_name, ast.NameConstant(None)],
                        keywords=[],
                    )
                ),
            ]


def _rewrite_function(f):
    """
    Rewrite a function so that any if/else branches are intercepted and their behavior can be
    overridden. This is accomplished using 3 steps:

    1. Get the source of the function and then rewrite the AST using _IfTransformer
    2. Do some small fixups to the tree to make sure
        a) the function doesn't have the same name, and
        b) the decorator isn't called recursively on the transformed function as well
    3. Bring the variables from the call site back into scope

    :param f: Function to rewrite
    :return: Rewritten function
    """
    source = inspect.getsource(f)
    tree = ast.parse(source)
    _IfTransformer().visit(tree)

    ast.fix_missing_locations(tree)
    tree.body[0].name = f.__name__ + "_patched"
    tree.body[0].decorator_list = []

    compiled = compile(tree, filename="<ast>", mode="exec")
    # The first f_back here gets to the body of magicquil() and the second f_back gets to the
    # user's call site which is what we want. If we didn't add these manually to the globals it
    # wouldn't be possible to call other @magicquil functions from within a @magicquil function.
    prev_globals = inspect.currentframe().f_back.f_back.f_globals
    # For reasons I don't quite understand it's critical to add locals() here otherwise the
    # function will disappear and we won't be able to return it below
    exec(compiled, {**prev_globals, **globals()}, locals())
    return locals()[f.__name__ + "_patched"]


def magicquil(f):
    """
    Decorator to enable a more convenient syntax for writing quil programs. With this decorator
    there is no need to keep track of a Program object and regular Python if/else branches can be
    used for classical control flow.

    Example usage:

    .. code-block:: python

        @magicquil
        def fast_reset(q1):
            reg1 = MEASURE(q1, None)
            if reg1:
                X(q1)
            else:
                I(q1)

        my_program = fast_reset(0)  # this will be a Program object
    """
    rewritten_function = _rewrite_function(f)

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if _program_context.get(None) is not None:
            rewritten_function(*args, **kwargs)
            program = _program_context.get()
        else:
            token = _program_context.set(Program())
            rewritten_function(*args, **kwargs)
            program = _program_context.get()
            _program_context.reset(token)
        return program

    return wrapper



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import logging
import sys
from types import TracebackType
from typing import Callable, Type

from pyquil.api._logger import logger


def exception_handler(
    exception_type: Type[BaseException],
    exception: BaseException,
    traceback: TracebackType,
    debug_hook: Callable[..., None] = sys.excepthook,
) -> None:
    """
    This allows us to suppress tracebacks for UserMessageError outside of debug mode
      by overriding the default exception handler.
    """
    if logger.level > logging.DEBUG and exception_type is UserMessageError:
        exception.__traceback__ = None
    debug_hook(exception_type, exception, traceback)


sys.excepthook = exception_handler  # type: ignore


class ApiError(RuntimeError):
    def __init__(self, server_status: str, explanation: str):
        super(ApiError, self).__init__(self, server_status)
        self.server_status = server_status
        self.explanation = explanation

    def __repr__(self) -> str:
        return repr(str(self))

    def __str__(self) -> str:
        return "{}\n{}".format(self.server_status, self.explanation)


class QVMError(ApiError):
    def __init__(self, server_status: str):
        explanation = """
The QVM returned the above error. This could be due to a bug in the server or a
bug in your code. If you suspect this to be a bug in pyQuil or Rigetti Forest,
then please describe the problem in a GitHub issue at:
    https://github.com/rigetti/pyquil/issues"""
        super(QVMError, self).__init__(server_status, explanation)


class QUILCError(ApiError):
    def __init__(self, server_status: str):
        explanation = """
QUILC returned the above error. This could be due to a bug in the server or a
bug in your code. If you suspect this to be a bug in pyQuil or Rigetti Forest,
then please describe the problem in a GitHub issue at:
    https://github.com/rigetti/pyquil/issues"""
        super(QUILCError, self).__init__(server_status, explanation)


class TooManyQubitsError(ApiError):
    def __init__(self, server_status: str):
        explanation = """
You requested too many qubits on the QVM. More qubits are available when you use
the queue. Pass the use_queue parameter to QVMConnection to enable additional
qubits (however, each program will take longer to run). For example:

    qvm = QVMConnection(use_queue=True)
    qvm.run(twenty_qubit_program)

See https://go.rigetti.com/connections for more info."""
        super(TooManyQubitsError, self).__init__(server_status, explanation)


class UserMessageError(Exception):
    """
    A special class of error which only displays its traceback when the program
      is run in debug mode (eg, with `LOG_LEVEL=DEBUG`).

    The purpose of this is to improve the user experience, reducing noise in
      the case of errors for which the cause is known.
    """

    def __init__(self, message: str):
        if logger.level <= logging.DEBUG:
            super().__init__(message)
        else:
            self.message = message

    def __str__(self) -> str:
        if logger.level <= logging.DEBUG:
            return super(UserMessageError, self).__str__()
        else:
            return f"ERROR: {self.message}"

    def __repr__(self) -> str:
        if logger.level <= logging.DEBUG:
            return super(UserMessageError, self).__repr__()
        else:
            return f"UserMessageError: {str(self)}"


class UnknownApiError(ApiError):
    def __init__(self, server_status: str):
        explanation = """
The server has failed to return a proper response. Please describe the problem
and copy the above message into a GitHub issue at:
    https://github.com/rigetti/pyquil/issues"""
        super(UnknownApiError, self).__init__(server_status, explanation)


error_mapping = {
    "quilc_error": QUILCError,
    "qvm_error": QVMError,
}



from pyquil.experiment._calibration import CalibrationMethod
from pyquil.experiment._group import (
    merge_disjoint_experiments,
    get_results_by_qubit_groups,
    group_settings,
)
from pyquil.experiment._main import (
    Experiment,
    OperatorEncoder,
    TomographyExperiment,
    read_json,
    to_json,
)
from pyquil.experiment._memory import merge_memory_map_lists
from pyquil.experiment._result import (
    ExperimentResult,
    bitstrings_to_expectations,
    correct_experiment_result,
    ratio_variance,
)
from pyquil.experiment._setting import (
    _OneQState,
    _pauli_to_product_state,
    ExperimentSetting,
    SIC0,
    SIC1,
    SIC2,
    SIC3,
    TensorProductState,
    minusX,
    minusY,
    minusZ,
    plusX,
    plusY,
    plusZ,
    zeros_state,
)
from pyquil.experiment._symmetrization import SymmetrizationLevel



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import uuid
import warnings
from collections import defaultdict
from typing import Any, Dict, List, Optional, Union, cast

import numpy as np
from rpcq.messages import QPURequest, ParameterAref, ParameterSpec

from pyquil.api import Client, QuantumExecutable, EncryptedProgram
from pyquil.api._error_reporting import _record_call
from pyquil.api._qam import QAM
from pyquil.quilatom import (
    MemoryReference,
    BinaryExp,
    Function,
    Parameter,
    ExpressionDesignator,
)


def decode_buffer(buffer: Dict[str, Any]) -> np.ndarray:
    """
    Translate a DataBuffer into a numpy array.

    :param buffer: Dictionary with 'data' byte array, 'dtype', and 'shape' fields
    :return: NumPy array of decoded data
    """
    buf = np.frombuffer(buffer["data"], dtype=buffer["dtype"])
    return buf.reshape(buffer["shape"])  # type: ignore


def _extract_memory_regions(
    memory_descriptors: Dict[str, ParameterSpec],
    ro_sources: Dict[MemoryReference, str],
    buffers: Dict[str, np.ndarray],
) -> Dict[str, np.ndarray]:

    # hack to extract num_shots indirectly from the shape of the returned data
    first, *rest = buffers.values()
    num_shots = first.shape[0]

    def alloc(spec: ParameterSpec) -> np.ndarray:
        dtype = {
            "BIT": np.int64,
            "INTEGER": np.int64,
            "REAL": np.float64,
            "FLOAT": np.float64,
        }
        try:
            return np.ndarray((num_shots, spec.length), dtype=dtype[spec.type])
        except KeyError:
            raise ValueError(f"Unexpected memory type {spec.type}.")

    regions: Dict[str, np.ndarray] = {}

    for mref, key in ro_sources.items():
        # Translation sometimes introduces ro_sources that the user didn't ask for.
        # That's fine, we just ignore them.
        if mref.name not in memory_descriptors:
            continue
        elif mref.name not in regions:
            regions[mref.name] = alloc(memory_descriptors[mref.name])

        buf = buffers[key]
        if buf.ndim == 1:
            buf = buf.reshape((num_shots, 1))

        if np.iscomplexobj(buf):
            buf = np.column_stack((buf.real, buf.imag))
        _, width = buf.shape

        end = mref.offset + width
        region_width = memory_descriptors[mref.name].length
        if end > region_width:
            raise ValueError(
                f"Attempted to fill {mref.name}[{mref.offset}, {end})"
                f"but the declared region has width {region_width}."
            )

        regions[mref.name][:, mref.offset : end] = buf

    return regions


class QPU(QAM):
    @_record_call
    def __init__(self, *, quantum_processor_id: str, client: Optional[Client] = None, priority: int = 1) -> None:
        """
        A connection to the QPU.

        :param quantum_processor_id: Processor to run against.
        :param client: Optional QCS client. If none is provided, a default client will be created.
        :param priority: The priority with which to insert jobs into the QPU queue. Lower
                         integers correspond to higher priority.
        """
        self.quantum_processor_id = quantum_processor_id
        self.priority = priority
        self._last_results: Dict[str, np.ndarray] = {}

        super().__init__(client)

    def get_version_info(self) -> Dict[str, Any]:
        """
        Return version information for this QPU's execution engine and its dependencies.

        :return: Dictionary of version information.
        """
        return cast(
            Dict[str, Any],
            self._client.processor_rpcq_request(self.quantum_processor_id, "get_version_info"),
        )

    @_record_call
    def load(self, executable: QuantumExecutable) -> "QPU":
        """
        Initialize a QAM into a fresh state. Load the executable and parse the expressions
        in the recalculation table (if any) into pyQuil Expression objects.

        :param executable: Load a compiled executable onto the QAM.
        """
        if not isinstance(executable, EncryptedProgram):
            raise TypeError(
                "`executable` argument must be an `EncryptedProgram`. Make "
                "sure you have explicitly compiled your program via `qc.compile` "
                "or `qc.compiler.native_quil_to_executable(...)` for more "
                "fine-grained control."
            )

        super().load(executable)
        return self

    @_record_call
    def run(self, run_priority: Optional[int] = None) -> "QPU":
        """
        Run a pyquil program on the QPU.

        This formats the classified data from the QPU server by stacking measured bits into
        an array of shape (trials, classical_addresses). The mapping of qubit to
        classical address is backed out from MEASURE instructions in the program, so
        only do measurements where there is a 1-to-1 mapping between qubits and classical
        addresses. If no MEASURE instructions are present in the program, a 0-by-0 array is
        returned.

        :param run_priority: The priority with which to insert jobs into the QPU queue. Lower
                             integers correspond to higher priority. If not specified, the QPU
                             object's default priority is used.
        :return: The QPU object itself.
        """
        super().run()
        assert isinstance(self.executable, EncryptedProgram)

        request = QPURequest(
            program=self.executable.program,
            patch_values=self._build_patch_values(),
            id=str(uuid.uuid4()),
        )

        job_priority = run_priority if run_priority is not None else self.priority
        job_id = self._client.processor_rpcq_request(
            self.quantum_processor_id,
            "execute_qpu_request",
            request,
            priority=job_priority,
        )
        results = self._get_buffers(job_id)
        ro_sources = self.executable.ro_sources

        self._memory_results = defaultdict(lambda: None)
        if results:
            extracted = _extract_memory_regions(self.executable.memory_descriptors, ro_sources, results)
            for name, array in extracted.items():
                self._memory_results[name] = array
        elif not ro_sources:
            warnings.warn(
                "You are running a QPU program with no MEASURE instructions. "
                "The result of this program will always be an empty array. Are "
                "you sure you didn't mean to measure some of your qubits?"
            )
            self._memory_results["ro"] = np.zeros((0, 0), dtype=np.int64)

        self._last_results = results

        return self

    def _get_buffers(self, job_id: str) -> Dict[str, np.ndarray]:
        """
        Return the decoded result buffers for particular job_id.

        :param job_id: Unique identifier for the job in question
        :return: Decoded buffers or throw an error
        """
        buffers = self._client.processor_rpcq_request(self.quantum_processor_id, "get_buffers", job_id, wait=True)
        return {k: decode_buffer(v) for k, v in buffers.items()}

    def _build_patch_values(self) -> Dict[str, List[Union[int, float]]]:
        patch_values = {}

        # Now that we are about to run, we have to resolve any gate parameter arithmetic that was
        # saved in the executable's recalculation table, and add those values to the variables shim
        self._update_variables_shim_with_recalculation_table()

        # Initialize our patch table
        assert isinstance(self.executable, EncryptedProgram)
        recalculation_table = self.executable.recalculation_table
        memory_ref_names = list(set(mr.name for mr in recalculation_table.keys()))
        if memory_ref_names:
            assert len(memory_ref_names) == 1, (
                "We expected only one declared memory region for "
                "the gate parameter arithmetic replacement references."
            )
            memory_reference_name = memory_ref_names[0]
            patch_values[memory_reference_name] = [0.0] * len(recalculation_table)

        for name, spec in self.executable.memory_descriptors.items():
            # NOTE: right now we fake reading out measurement values into classical memory
            # hence we omit them here from the patch table.
            if any(name == mref.name for mref in self.executable.ro_sources):
                continue
            initial_value = 0.0 if spec.type == "REAL" else 0
            patch_values[name] = [initial_value] * spec.length

        # Fill in our patch table
        for k, v in self._variables_shim.items():
            patch_values[k.name][k.index] = v

        return patch_values

    def _update_variables_shim_with_recalculation_table(self) -> None:
        """
        Update self._variables_shim with the final values to be patched into the gate parameters,
        according to the arithmetic expressions in the original program.

        For example::

            DECLARE theta REAL
            DECLARE beta REAL
            RZ(3 * theta) 0
            RZ(beta+theta) 0

        gets translated to::

            DECLARE theta REAL
            DECLARE __P REAL[2]
            RZ(__P[0]) 0
            RZ(__P[1]) 0

        and the recalculation table will contain::

            {
                ParameterAref('__P', 0): Mul(3.0, <MemoryReference theta[0]>),
                ParameterAref('__P', 1): Add(<MemoryReference beta[0]>, <MemoryReference theta[0]>)
            }

        Let's say we've made the following two function calls:

        .. code-block:: python

            qpu.write_memory(region_name='theta', value=0.5)
            qpu.write_memory(region_name='beta', value=0.1)

        After executing this function, our self.variables_shim in the above example would contain
        the following:

        .. code-block:: python

            {
                ParameterAref('theta', 0): 0.5,
                ParameterAref('beta', 0): 0.1,
                ParameterAref('__P', 0): 1.5,       # (3.0) * theta[0]
                ParameterAref('__P', 1): 0.6        # beta[0] + theta[0]
            }

        Once the _variables_shim is filled, execution continues as with regular binary patching.
        """
        assert isinstance(self.executable, EncryptedProgram)
        for mref, expression in self.executable.recalculation_table.items():
            # Replace the user-declared memory references with any values the user has written,
            # coerced to a float because that is how we declared it.
            self._variables_shim[mref] = float(self._resolve_memory_references(expression))

    def _resolve_memory_references(self, expression: ExpressionDesignator) -> Union[float, int]:
        """
        Traverse the given Expression, and replace any Memory References with whatever values
        have been so far provided by the user for those memory spaces. Declared memory defaults
        to zero.

        :param expression: an Expression
        """
        if isinstance(expression, BinaryExp):
            left = self._resolve_memory_references(expression.op1)
            right = self._resolve_memory_references(expression.op2)
            return cast(Union[float, int], expression.fn(left, right))
        elif isinstance(expression, Function):
            return cast(
                Union[float, int],
                expression.fn(self._resolve_memory_references(expression.expression)),
            )
        elif isinstance(expression, Parameter):
            raise ValueError(f"Unexpected Parameter in gate expression: {expression}")
        elif isinstance(expression, (float, int)):
            return expression
        elif isinstance(expression, MemoryReference):
            return self._variables_shim.get(ParameterAref(name=expression.name, index=expression.offset), 0)
        else:
            raise ValueError(f"Unexpected expression in gate parameter: {expression}")



import math

import pytest
from _pytest.monkeypatch import MonkeyPatch

from pyquil import Program
from pyquil.api import Client
from pyquil.api._compiler import QPUCompiler
from pyquil.device import QCSDevice
from pyquil.gates import RX, MEASURE, RZ
from pyquil.quilatom import FormalArgument
from pyquil.quilbase import DefCalibration


# TODO(andrew): tests
# QPUCompiler
# QVMCompiler


def simple_program():
    program = Program()
    readout = program.declare("ro", "BIT", 3)
    program += RX(math.pi / 2, 0)
    program += MEASURE(0, readout[0])
    return program


def test_invalid_protocol(qcs_aspen8_device: QCSDevice, monkeypatch: MonkeyPatch):
    monkeypatch.setenv("QCS_SETTINGS_APPLICATIONS_PYQUIL_QUILC_URL", "not-http-or-tcp://example.com")
    client = Client()

    with pytest.raises(
        ValueError,
        match="Expected compiler URL 'not-http-or-tcp://example.com' to start with 'tcp://'",
    ):
        QPUCompiler(
            quantum_processor_id=qcs_aspen8_device.quantum_processor_id,
            device=qcs_aspen8_device,
            client=client,
        )


def test_compile_with_quilt_calibrations(compiler: QPUCompiler):
    program = simple_program()
    q = FormalArgument("q")
    defn = DefCalibration("H", [], [q], [RZ(math.pi / 2, q), RX(math.pi / 2, q), RZ(math.pi / 2, q)])
    cals = [defn]
    program._calibrations = cals
    # this should more or less pass through
    compilation_result = compiler.quil_to_native_quil(program, protoquil=True)
    assert compilation_result.calibrations == cals
    assert program.calibrations == cals
    assert compilation_result == program



import numpy as np
import pytest

from pyquil import Program
from pyquil.api import QVM, Client
from pyquil.api._errors import QVMError
from pyquil.api._qvm import validate_noise_probabilities, validate_qubit_list, prepare_register_list
from pyquil.gates import MEASURE, X
from pyquil.quilbase import Declare, MemoryReference


def test_qvm__default_client(client):
    qvm = QVM(client=client)
    p = Program(Declare("ro", "BIT"), X(0), MEASURE(0, MemoryReference("ro")))
    qvm.load(p.wrap_in_numshots_loop(1000))
    qvm.run()
    qvm.wait()
    bitstrings = qvm.read_memory(region_name="ro")
    assert bitstrings.shape == (1000, 1)


def test_qvm_run_pqer(client: Client):
    qvm = QVM(client=client, gate_noise=(0.01, 0.01, 0.01))
    p = Program(Declare("ro", "BIT"), X(0), MEASURE(0, MemoryReference("ro")))
    qvm.load(p.wrap_in_numshots_loop(1000))
    qvm.run()
    qvm.wait()
    bitstrings = qvm.read_memory(region_name="ro")
    assert bitstrings.shape == (1000, 1)
    assert np.mean(bitstrings) > 0.8


def test_qvm_run_just_program(client: Client):
    qvm = QVM(client=client, gate_noise=(0.01, 0.01, 0.01))
    p = Program(Declare("ro", "BIT"), X(0), MEASURE(0, MemoryReference("ro")))
    qvm.load(p.wrap_in_numshots_loop(1000))
    qvm.run()
    qvm.wait()
    bitstrings = qvm.read_memory(region_name="ro")
    assert bitstrings.shape == (1000, 1)
    assert np.mean(bitstrings) > 0.8


def test_qvm_run_only_pqer(client: Client):
    qvm = QVM(client=client, gate_noise=(0.01, 0.01, 0.01))
    p = Program(Declare("ro", "BIT"), X(0), MEASURE(0, MemoryReference("ro")))

    qvm.load(p.wrap_in_numshots_loop(1000))
    qvm.run()
    qvm.wait()
    bitstrings = qvm.read_memory(region_name="ro")
    assert bitstrings.shape == (1000, 1)
    assert np.mean(bitstrings) > 0.8


def test_qvm_run_region_declared_and_measured(client: Client):
    qvm = QVM(client=client)
    p = Program(Declare("reg", "BIT"), X(0), MEASURE(0, MemoryReference("reg")))
    qvm.load(p.wrap_in_numshots_loop(100)).run().wait()
    bitstrings = qvm.read_memory(region_name="reg")
    assert bitstrings.shape == (100, 1)


def test_qvm_run_region_declared_not_measured(client: Client):
    qvm = QVM(client=client)
    p = Program(Declare("reg", "BIT"), X(0))
    qvm.load(p.wrap_in_numshots_loop(100)).run().wait()
    bitstrings = qvm.read_memory(region_name="reg")
    assert bitstrings.shape == (100, 0)


# For backwards compatibility, we support omitting the declaration for "ro" specifically
# TODO(andrew): we should remove this lenient behavior
def test_qvm_run_region_not_declared_is_measured_ro(client: Client):
    qvm = QVM(client=client)
    p = Program(X(0), MEASURE(0, MemoryReference("ro")))
    qvm.load(p.wrap_in_numshots_loop(100)).run().wait()
    bitstrings = qvm.read_memory(region_name="ro")
    assert bitstrings.shape == (100, 1)


def test_qvm_run_region_not_declared_is_measured_non_ro(client: Client):
    qvm = QVM(client=client)
    p = Program(X(0), MEASURE(0, MemoryReference("reg")))

    with pytest.raises(QVMError, match='Bad memory region name "reg" in MEASURE'):
        qvm.load(p).run().wait()


def test_qvm_run_region_not_declared_not_measured_ro(client: Client):
    qvm = QVM(client=client)
    p = Program(X(0))
    qvm.load(p.wrap_in_numshots_loop(100)).run().wait()
    bitstrings = qvm.read_memory(region_name="ro")
    assert bitstrings.shape == (100, 0)


def test_qvm_run_region_not_declared_not_measured_non_ro(client: Client):
    qvm = QVM(client=client)
    p = Program(X(0))
    qvm.load(p.wrap_in_numshots_loop(100)).run().wait()
    assert qvm.read_memory(region_name="reg") is None


def test_qvm_version(client: Client):
    qvm = QVM(client=client)
    version = qvm.get_version_info()

    def is_a_version_string(version_string: str):
        parts = version_string.split(".")
        try:
            map(int, parts)
        except ValueError:
            return False
        return True

    assert is_a_version_string(version)


def test_validate_noise_probabilities():
    with pytest.raises(TypeError, match="noise_parameter must be a tuple"):
        validate_noise_probabilities(1)
    with pytest.raises(TypeError, match="noise_parameter values should all be floats"):
        validate_noise_probabilities(("a", "b", "c"))
    with pytest.raises(ValueError, match="noise_parameter tuple must be of length 3"):
        validate_noise_probabilities((0.0, 0.0, 0.0, 0.0))
    with pytest.raises(
        ValueError,
        match="sum of entries in noise_parameter must be between 0 and 1 \\(inclusive\\)",
    ):
        validate_noise_probabilities((0.5, 0.5, 0.5))
    with pytest.raises(ValueError, match="noise_parameter values should all be non-negative"):
        validate_noise_probabilities((-0.5, -0.5, 1.0))


def test_validate_qubit_list():
    with pytest.raises(TypeError):
        validate_qubit_list([-1, 1])
    with pytest.raises(TypeError):
        validate_qubit_list(["a", 0], 1)


def test_prepare_register_list():
    with pytest.raises(TypeError):
        prepare_register_list({"ro": [-1, 1]})



import itertools
import pytest

from pyquil.experiment._main import Experiment
from pyquil.experiment._result import ExperimentResult
from pyquil.experiment._setting import (
    ExperimentSetting,
    plusX,
    minusX,
    plusZ,
    minusZ,
    plusY,
    TensorProductState,
    zeros_state,
    _pauli_to_product_state,
)
from pyquil.experiment._group import (
    get_results_by_qubit_groups,
    merge_disjoint_experiments,
    group_settings,
    _max_weight_operator,
    _max_weight_state,
    _max_tpb_overlap,
)
from pyquil.gates import X, Z
from pyquil.paulis import sZ, sX, sY, sI, PauliTerm
from pyquil import Program
from pyquil.gates import H


def test_results_by_qubit_groups():
    er1 = ExperimentResult(
        setting=ExperimentSetting(plusX(0), sZ(0)),
        expectation=0.0,
        std_err=0.0,
        total_counts=1,
    )

    er2 = ExperimentResult(
        setting=ExperimentSetting(plusX(0), sZ(1)),
        expectation=0.0,
        std_err=0.0,
        total_counts=1,
    )

    er3 = ExperimentResult(
        setting=ExperimentSetting(plusX(0), sX(0) * sZ(1)),
        expectation=0.0,
        std_err=0.0,
        total_counts=1,
    )

    er4 = ExperimentResult(
        setting=ExperimentSetting(plusX(0), sX(0) * sZ(2)),
        expectation=0.0,
        std_err=0.0,
        total_counts=1,
    )
    groups = [(0,), (1,), (2, 0)]
    res_by_group = get_results_by_qubit_groups([er1, er2, er3, er4], groups)

    assert res_by_group == {(0,): [er1], (1,): [er2], (0, 2): [er1, er4]}


def test_merge_disjoint_experiments():
    sett1 = ExperimentSetting(TensorProductState(), sX(0) * sY(1))
    sett2 = ExperimentSetting(plusZ(1), sY(1))
    sett3 = ExperimentSetting(plusZ(0), sX(0))
    sett4 = ExperimentSetting(minusX(1), sY(1))
    sett5 = ExperimentSetting(TensorProductState(), sZ(2))

    expt1 = Experiment(settings=[sett1, sett2], program=Program(X(1)))
    expt2 = Experiment(settings=[sett3, sett4], program=Program(Z(0)))
    expt3 = Experiment(settings=[sett5], program=Program())

    merged_expt = merge_disjoint_experiments([expt1, expt2, expt3])
    assert len(merged_expt) == 2


@pytest.fixture(params=["clique-removal", "greedy"])
def grouping_method(request):
    return request.param


def test_expt_settings_share_ntpb():
    expts = [
        [
            ExperimentSetting(zeros_state([0, 1]), sX(0) * sI(1)),
            ExperimentSetting(zeros_state([0, 1]), sI(0) * sX(1)),
        ],
        [
            ExperimentSetting(zeros_state([0, 1]), sZ(0) * sI(1)),
            ExperimentSetting(zeros_state([0, 1]), sI(0) * sZ(1)),
        ],
    ]
    for group in expts:
        for e1, e2 in itertools.combinations(group, 2):
            assert _max_weight_state([e1.in_state, e2.in_state]) is not None
            assert _max_weight_operator([e1.out_operator, e2.out_operator]) is not None


def test_group_experiments(grouping_method):
    expts = [  # cf above, I removed the inner nesting. Still grouped visually
        ExperimentSetting(TensorProductState(), sX(0) * sI(1)),
        ExperimentSetting(TensorProductState(), sI(0) * sX(1)),
        ExperimentSetting(TensorProductState(), sZ(0) * sI(1)),
        ExperimentSetting(TensorProductState(), sI(0) * sZ(1)),
    ]
    suite = Experiment(expts, Program())
    grouped_suite = group_settings(suite, method=grouping_method)
    assert len(suite) == 4
    assert len(grouped_suite) == 2


def test_max_weight_operator_1():
    pauli_terms = [sZ(0), sX(1) * sZ(0), sY(2) * sX(1)]
    assert _max_weight_operator(pauli_terms) == sY(2) * sX(1) * sZ(0)


def test_max_weight_operator_2():
    pauli_terms = [sZ(0), sX(1) * sZ(0), sY(2) * sX(1), sZ(5) * sI(3)]
    assert _max_weight_operator(pauli_terms) == sZ(5) * sY(2) * sX(1) * sZ(0)


def test_max_weight_operator_3():
    pauli_terms = [sZ(0) * sX(5), sX(1) * sZ(0), sY(2) * sX(1), sZ(5) * sI(3)]
    assert _max_weight_operator(pauli_terms) is None


def test_max_weight_operator_misc():
    assert _max_weight_operator([sZ(0), sZ(0) * sZ(1)]) is not None
    assert _max_weight_operator([sX(5), sZ(4)]) is not None
    assert _max_weight_operator([sX(0), sY(0) * sZ(2)]) is None

    x_term = sX(0) * sX(1)
    z1_term = sZ(1)
    z0_term = sZ(0)
    z0z1_term = sZ(0) * sZ(1)
    assert _max_weight_operator([x_term, z1_term]) is None
    assert _max_weight_operator([z0z1_term, x_term]) is None

    assert _max_weight_operator([z1_term, z0_term]) is not None
    assert _max_weight_operator([z0z1_term, z0_term]) is not None
    assert _max_weight_operator([z0z1_term, z1_term]) is not None
    assert _max_weight_operator([z0z1_term, sI(1)]) is not None
    assert _max_weight_operator([z0z1_term, sI(2)]) is not None
    assert _max_weight_operator([z0z1_term, sX(5) * sZ(7)]) is not None

    xxxx_terms = (
        sX(1) * sX(2) + sX(2) + sX(3) * sX(4) + sX(4) + sX(1) * sX(3) * sX(4) + sX(1) * sX(4) + sX(1) * sX(2) * sX(3)
    )
    true_term = sX(1) * sX(2) * sX(3) * sX(4)
    assert _max_weight_operator(xxxx_terms.terms) == true_term

    zzzz_terms = sZ(1) * sZ(2) + sZ(3) * sZ(4) + sZ(1) * sZ(3) + sZ(1) * sZ(3) * sZ(4)
    assert _max_weight_operator(zzzz_terms.terms) == sZ(1) * sZ(2) * sZ(3) * sZ(4)

    pauli_terms = [sZ(0), sX(1) * sZ(0), sY(2) * sX(1), sZ(5) * sI(3)]
    assert _max_weight_operator(pauli_terms) == sZ(5) * sY(2) * sX(1) * sZ(0)


def test_max_weight_operator_4():
    # this last example illustrates that a pair of commuting operators
    # need not be diagonal in the same tpb
    assert _max_weight_operator([sX(1) * sZ(0), sZ(1) * sX(0)]) is None


def test_max_weight_state_1():
    states = [plusX(0) * plusZ(1), plusX(0), plusZ(1)]
    assert _max_weight_state(states) == states[0]


def test_max_weight_state_2():
    states = [plusX(1) * plusZ(0), plusX(0), plusZ(1)]
    assert _max_weight_state(states) is None


def test_max_weight_state_3():
    states = [plusX(0) * minusZ(1), plusX(0), minusZ(1)]
    assert _max_weight_state(states) == states[0]


def test_max_weight_state_4():
    states = [plusX(1) * minusZ(0), plusX(0), minusZ(1)]
    assert _max_weight_state(states) is None


def test_max_tpb_overlap_1():
    tomo_expt_settings = [
        ExperimentSetting(plusZ(1) * plusX(0), sY(2) * sY(1)),
        ExperimentSetting(plusX(2) * plusZ(1), sY(2) * sZ(0)),
    ]
    tomo_expt_program = Program(H(0), H(1), H(2))
    tomo_expt = Experiment(tomo_expt_settings, tomo_expt_program)
    expected_dict = {
        ExperimentSetting(plusX(0) * plusZ(1) * plusX(2), sZ(0) * sY(1) * sY(2)): [
            ExperimentSetting(plusZ(1) * plusX(0), sY(2) * sY(1)),
            ExperimentSetting(plusX(2) * plusZ(1), sY(2) * sZ(0)),
        ]
    }
    assert expected_dict == _max_tpb_overlap(tomo_expt)


def test_max_tpb_overlap_2():
    expt_setting = ExperimentSetting(
        _pauli_to_product_state(PauliTerm.from_compact_str("(1+0j)*Z7Y8Z1Y4Z2Y5Y0X6")),
        PauliTerm.from_compact_str("(1+0j)*Z4X8Y5X3Y7Y1"),
    )
    p = Program(H(0), H(1), H(2))
    tomo_expt = Experiment([expt_setting], p)
    expected_dict = {expt_setting: [expt_setting]}
    assert expected_dict == _max_tpb_overlap(tomo_expt)


def test_max_tpb_overlap_3():
    # add another ExperimentSetting to the above
    expt_setting = ExperimentSetting(
        _pauli_to_product_state(PauliTerm.from_compact_str("(1+0j)*Z7Y8Z1Y4Z2Y5Y0X6")),
        PauliTerm.from_compact_str("(1+0j)*Z4X8Y5X3Y7Y1"),
    )
    expt_setting2 = ExperimentSetting(plusZ(7), sY(1))
    p = Program(H(0), H(1), H(2))
    tomo_expt2 = Experiment([expt_setting, expt_setting2], p)
    expected_dict2 = {expt_setting: [expt_setting, expt_setting2]}
    assert expected_dict2 == _max_tpb_overlap(tomo_expt2)


def test_group_experiments_greedy():
    ungrouped_tomo_expt = Experiment(
        [
            [
                ExperimentSetting(
                    _pauli_to_product_state(PauliTerm.from_compact_str("(1+0j)*Z7Y8Z1Y4Z2Y5Y0X6")),
                    PauliTerm.from_compact_str("(1+0j)*Z4X8Y5X3Y7Y1"),
                )
            ],
            [ExperimentSetting(plusZ(7), sY(1))],
        ],
        program=Program(H(0), H(1), H(2)),
    )
    grouped_tomo_expt = group_settings(ungrouped_tomo_expt, method="greedy")
    expected_grouped_tomo_expt = Experiment(
        [
            [
                ExperimentSetting(
                    TensorProductState.from_str("Z0_7 * Y0_8 * Z0_1 * Y0_4 * Z0_2 * Y0_5 * Y0_0 * X0_6"),
                    PauliTerm.from_compact_str("(1+0j)*Z4X8Y5X3Y7Y1"),
                ),
                ExperimentSetting(plusZ(7), sY(1)),
            ]
        ],
        program=Program(H(0), H(1), H(2)),
    )
    assert grouped_tomo_expt == expected_grouped_tomo_expt


def test_expt_settings_diagonal_in_tpb():
    def _expt_settings_diagonal_in_tpb(es1: ExperimentSetting, es2: ExperimentSetting):
        """
        Extends the concept of being diagonal in the same tpb to ExperimentSettings, by
        determining if the pairs of in_states and out_operators are separately diagonal in the same
        tpb
        """
        max_weight_in = _max_weight_state([es1.in_state, es2.in_state])
        max_weight_out = _max_weight_operator([es1.out_operator, es2.out_operator])
        return max_weight_in is not None and max_weight_out is not None

    expt_setting1 = ExperimentSetting(plusZ(1) * plusX(0), sY(1) * sZ(0))
    expt_setting2 = ExperimentSetting(plusY(2) * plusZ(1), sZ(2) * sY(1))
    assert _expt_settings_diagonal_in_tpb(expt_setting1, expt_setting2)
    expt_setting3 = ExperimentSetting(plusX(2) * plusZ(1), sZ(2) * sY(1))
    expt_setting4 = ExperimentSetting(plusY(2) * plusZ(1), sX(2) * sY(1))
    assert not _expt_settings_diagonal_in_tpb(expt_setting2, expt_setting3)
    assert not _expt_settings_diagonal_in_tpb(expt_setting2, expt_setting4)



import numpy as np

from pyquil import Program
from pyquil.experiment._main import _remove_reset_from_program
from pyquil.experiment._program import (
    parameterized_readout_symmetrization,
    parameterized_single_qubit_measurement_basis,
    parameterized_single_qubit_state_preparation,
)
from pyquil.experiment import (
    ExperimentSetting,
    TensorProductState,
    Experiment,
    plusZ,
    read_json,
    to_json,
)
from pyquil.gates import MEASURE, RESET, X, Y
from pyquil.paulis import sI, sX, sY, sZ


EXPERIMENT_REPR = """
shots: 1
active reset: disabled
symmetrization: -1 (exhaustive)
calibration: 1 (plus_eigenstate)
program:
   X 0
   Y 1
"""


def test_tomo_experiment():
    expts = [
        ExperimentSetting(TensorProductState(), sX(0) * sY(1)),
        ExperimentSetting(plusZ(0), sZ(0)),
    ]

    suite = Experiment(settings=expts, program=Program(X(0), Y(1)))
    assert len(suite) == 2
    for e1, e2 in zip(expts, suite):
        # experiment suite puts in groups of length 1
        assert len(e2) == 1
        e2 = e2[0]
        assert e1 == e2
    prog_str = str(suite).splitlines()[3:5]
    assert prog_str == EXPERIMENT_REPR.splitlines()[4:6]


def test_tomo_experiment_pre_grouped():
    expts = [
        [
            ExperimentSetting(TensorProductState(), sX(0) * sI(1)),
            ExperimentSetting(TensorProductState(), sI(0) * sX(1)),
        ],
        [
            ExperimentSetting(TensorProductState(), sZ(0) * sI(1)),
            ExperimentSetting(TensorProductState(), sI(0) * sZ(1)),
        ],
    ]

    suite = Experiment(settings=expts, program=Program(X(0), Y(1)))
    assert len(suite) == 2  # number of groups
    for es1, es2 in zip(expts, suite):
        for e1, e2 in zip(es1, es2):
            assert e1 == e2
    prog_str = str(suite).splitlines()[3:5]
    assert prog_str == EXPERIMENT_REPR.splitlines()[4:6]


def test_tomo_experiment_empty():
    suite = Experiment([], program=Program(X(0)))
    assert len(suite) == 0
    assert str(suite.program) == "X 0\n"


def test_experiment_deser(tmpdir):
    expts = [
        [
            ExperimentSetting(TensorProductState(), sX(0) * sI(1)),
            ExperimentSetting(TensorProductState(), sI(0) * sX(1)),
        ],
        [
            ExperimentSetting(TensorProductState(), sZ(0) * sI(1)),
            ExperimentSetting(TensorProductState(), sI(0) * sZ(1)),
        ],
    ]

    suite = Experiment(settings=expts, program=Program(X(0), Y(1)))
    to_json(f"{tmpdir}/suite.json", suite)
    suite2 = read_json(f"{tmpdir}/suite.json")
    assert suite == suite2


DEFGATE_X = """
DEFGATE XGATE:
    0, 1
    1, 0
"""


TRIMMED_PROG = """
DEFGATE XGATE:
    0, 1
    1, 0

X 0
"""


def test_remove_reset_from_program():
    p = Program(DEFGATE_X)
    p += RESET()
    p += X(0)
    new_p = _remove_reset_from_program(p)
    assert "\n" + new_p.out() == TRIMMED_PROG


def test_generate_experiment_program():
    # simplest example
    p = Program()
    s = ExperimentSetting(in_state=sZ(0), out_operator=sZ(0))
    e = Experiment(settings=[s], program=p, symmetrization=0)
    exp = e.generate_experiment_program()
    test_exp = Program()
    ro = test_exp.declare("ro", "BIT")
    test_exp += MEASURE(0, ro[0])
    assert exp.out() == test_exp.out()
    assert exp.num_shots == 1

    # 2Q exhaustive symmetrization
    p = Program()
    s = ExperimentSetting(in_state=sZ(0) * sZ(1), out_operator=sZ(0) * sZ(1))
    e = Experiment(settings=[s], program=p)
    exp = e.generate_experiment_program()
    test_exp = Program()
    test_exp += parameterized_readout_symmetrization([0, 1])
    ro = test_exp.declare("ro", "BIT", 2)
    test_exp += MEASURE(0, ro[0])
    test_exp += MEASURE(1, ro[1])
    assert exp.out() == test_exp.out()
    assert exp.num_shots == 1

    # add shots
    p = Program()
    p.wrap_in_numshots_loop(1000)
    s = ExperimentSetting(in_state=sZ(0), out_operator=sZ(0))
    e = Experiment(settings=[s], program=p, symmetrization=0)
    exp = e.generate_experiment_program()
    test_exp = Program()
    ro = test_exp.declare("ro", "BIT")
    test_exp += MEASURE(0, ro[0])
    assert exp.out() == test_exp.out()
    assert exp.num_shots == 1000

    # active reset
    p = Program()
    p += RESET()
    s = ExperimentSetting(in_state=sZ(0), out_operator=sZ(0))
    e = Experiment(settings=[s], program=p, symmetrization=0)
    exp = e.generate_experiment_program()
    test_exp = Program()
    test_exp += RESET()
    ro = test_exp.declare("ro", "BIT")
    test_exp += MEASURE(0, ro[0])
    assert exp.out() == test_exp.out()
    assert exp.num_shots == 1

    # state preparation and measurement
    p = Program()
    s = ExperimentSetting(in_state=sY(0), out_operator=sX(0))
    e = Experiment(settings=[s], program=p, symmetrization=0)
    exp = e.generate_experiment_program()
    test_exp = Program()
    test_exp += parameterized_single_qubit_state_preparation([0])
    test_exp += parameterized_single_qubit_measurement_basis([0])
    ro = test_exp.declare("ro", "BIT")
    test_exp += MEASURE(0, ro[0])
    assert exp.out() == test_exp.out()
    assert exp.num_shots == 1

    # multi-qubit state preparation and measurement
    p = Program()
    s = ExperimentSetting(in_state=sZ(0) * sY(1), out_operator=sZ(0) * sX(1))
    e = Experiment(settings=[s], program=p, symmetrization=0)
    exp = e.generate_experiment_program()
    test_exp = Program()
    test_exp += parameterized_single_qubit_state_preparation([0, 1])
    test_exp += parameterized_single_qubit_measurement_basis([0, 1])
    ro = test_exp.declare("ro", "BIT", 2)
    test_exp += MEASURE(0, ro[0])
    test_exp += MEASURE(1, ro[1])
    assert exp.out() == test_exp.out()
    assert exp.num_shots == 1


def test_build_experiment_setting_memory_map():
    p = Program()
    s = ExperimentSetting(in_state=sX(0), out_operator=sZ(0) * sY(1))
    e = Experiment(settings=[s], program=p)
    memory_map = e.build_setting_memory_map(s)
    assert memory_map == {
        "preparation_alpha": [0.0],
        "preparation_beta": [np.pi / 2],
        "preparation_gamma": [0.0],
        "measurement_alpha": [0.0, np.pi / 2],
        "measurement_beta": [0.0, np.pi / 2],
        "measurement_gamma": [0.0, -np.pi / 2],
    }


def test_build_symmetrization_memory_maps():
    p = Program()
    s = ExperimentSetting(in_state=sZ(0) * sZ(1), out_operator=sZ(0) * sZ(1))
    e = Experiment(settings=[s], program=p)
    memory_maps = [
        {"symmetrization": [0.0, 0.0]},
        {"symmetrization": [0.0, np.pi]},
        {"symmetrization": [np.pi, 0.0]},
        {"symmetrization": [np.pi, np.pi]},
    ]
    assert e.build_symmetrization_memory_maps([0, 1]) == memory_maps



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import logging
from typing import Dict, Optional, cast, List

from qcs_api_client.models import TranslateNativeQuilToEncryptedBinaryResponse
from qcs_api_client.models.get_quilt_calibrations_response import GetQuiltCalibrationsResponse
from qcs_api_client.models.translate_native_quil_to_encrypted_binary_request import (
    TranslateNativeQuilToEncryptedBinaryRequest,
)
from qcs_api_client.operations.sync import (
    translate_native_quil_to_encrypted_binary,
    get_quilt_calibrations,
)
from qcs_api_client.types import UNSET
from rpcq.messages import ParameterSpec

from pyquil.api import Client
from pyquil.api._error_reporting import _record_call
from pyquil.api._abstract_compiler import AbstractCompiler, QuantumExecutable, EncryptedProgram
from pyquil.api._rewrite_arithmetic import rewrite_arithmetic
from pyquil.device import AbstractDevice
from pyquil.parser import parse_program, parse
from pyquil.quil import Program
from pyquil.quilatom import MemoryReference, ExpressionDesignator
from pyquil.quilbase import Declare, Gate

_log = logging.getLogger(__name__)

PYQUIL_PROGRAM_PROPERTIES = ["native_quil_metadata", "num_shots"]


class QPUCompilerNotRunning(Exception):
    pass


def parse_mref(val: str) -> MemoryReference:
    """ Parse a memory reference from its string representation. """
    val = val.strip()
    try:
        if val[-1] == "]":
            name, offset = val.split("[")
            return MemoryReference(name, int(offset[:-1]))
        else:
            return MemoryReference(val)
    except Exception:
        raise ValueError(f"Unable to parse memory reference {val}.")


def _collect_memory_descriptors(program: Program) -> Dict[str, ParameterSpec]:
    """Collect Declare instructions that are important for building the patch table.

    :return: A dictionary of variable names to specs about the declared region.
    """

    return {
        instr.name: ParameterSpec(type=instr.memory_type, length=instr.memory_size)
        for instr in program
        if isinstance(instr, Declare)
    }


class QPUCompiler(AbstractCompiler):
    """
    Client to communicate with the compiler and translation service.
    """

    quantum_processor_id: str
    _calibration_program: Optional[Program] = None

    @_record_call
    def __init__(
        self,
        *,
        quantum_processor_id: str,
        device: AbstractDevice,
        client: Optional[Client] = None,
        timeout: float = 10,
    ) -> None:
        """
        Instantiate a new QPU compiler client.

        :param quantum_processor_id: Processor to target.
        :param device: PyQuil Device object to use as compilation target.
        :param client: Optional QCS client. If none is provided, a default client will be created.
        :param timeout: Number of seconds to wait for a response from the client.
        """
        super().__init__(device=device, client=client, timeout=timeout)
        self.quantum_processor_id = quantum_processor_id

    @_record_call
    def native_quil_to_executable(self, nq_program: Program) -> QuantumExecutable:
        arithmetic_response = rewrite_arithmetic(nq_program)
        request = TranslateNativeQuilToEncryptedBinaryRequest(
            quil=arithmetic_response.quil, num_shots=nq_program.num_shots
        )

        # TODO(andrew): timeout?
        response = cast(
            TranslateNativeQuilToEncryptedBinaryResponse,
            self._client.qcs_request(
                translate_native_quil_to_encrypted_binary,
                quantum_processor_id=self.quantum_processor_id,
                json_body=request,
            ),
        )

        ro_sources = cast(List[List[str]], [] if response.ro_sources == UNSET else response.ro_sources)

        def to_expression(rule: str) -> ExpressionDesignator:
            # We can only parse complete lines of Quil, so we wrap the arithmetic expression
            # in a valid Quil instruction to parse it.
            # TODO: This hack should be replaced after #687
            return cast(ExpressionDesignator, cast(Gate, parse(f"RZ({rule}) 0")[0]).params[0])

        return EncryptedProgram(
            program=response.program,
            memory_descriptors=_collect_memory_descriptors(nq_program),
            ro_sources={parse_mref(mref): source for mref, source in ro_sources},
            recalculation_table={
                mref: to_expression(rule) for mref, rule in arithmetic_response.recalculation_table.items()
            },
        )

    def _get_calibration_program(self) -> Program:
        response = cast(
            GetQuiltCalibrationsResponse,
            self._client.qcs_request(get_quilt_calibrations, quantum_processor_id=self.quantum_processor_id),
        )
        return parse_program(response.quilt)

    @property  # type: ignore
    @_record_call
    def calibration_program(self) -> Program:
        """
        Get the Quil-T calibration program associated with the underlying QPU.

        A calibration program contains a number of DEFCAL, DEFWAVEFORM, and
        DEFFRAME instructions. In sum, those instructions describe how a Quil-T
        program should be translated into analog instructions for execution on
        hardware.

        :returns: A Program object containing the calibration definitions."""
        if self._calibration_program is None:
            try:
                self._calibration_program = self._get_calibration_program()
            except Exception as ex:
                raise RuntimeError("Could not refresh calibrations") from ex

        return self._calibration_program

    @_record_call
    def reset(self) -> None:
        """
        Reset the state of the QPUCompiler.
        """
        super().reset()
        self._calibration_program = None


class QVMCompiler(AbstractCompiler):
    """
    Client to communicate with the compiler.
    """

    @_record_call
    def __init__(self, *, device: AbstractDevice, client: Optional[Client] = None, timeout: float = 10) -> None:
        """
        Client to communicate with compiler.

        :param device: PyQuil Device object to use as compilation target.
        :param client: Optional QCS client. If none is provided, a default client will be created.
        :param timeout: Number of seconds to wait for a response from the client.
        """
        super().__init__(device=device, client=client, timeout=timeout)

    @_record_call
    def native_quil_to_executable(self, nq_program: Program) -> QuantumExecutable:
        return nq_program



import numpy as np

from pyquil.experiment import plusX
from pyquil.experiment._result import (
    bitstrings_to_expectations,
    correct_experiment_result,
    ratio_variance,
    ExperimentResult,
    ExperimentSetting,
)
from pyquil.paulis import sZ


def test_bitstrings_to_expectations():
    bitstrings = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])

    assert np.allclose(bitstrings_to_expectations(bitstrings), np.array([[1, 1], [1, -1], [-1, 1], [-1, -1]]))
    assert np.allclose(
        bitstrings_to_expectations(bitstrings, joint_expectations=[[0]]),
        np.array([[1], [1], [-1], [-1]]),
    )
    assert np.allclose(
        bitstrings_to_expectations(bitstrings, joint_expectations=[[1]]),
        np.array([[1], [-1], [1], [-1]]),
    )
    assert np.allclose(
        bitstrings_to_expectations(bitstrings, joint_expectations=[[0, 1]]),
        np.array([[1], [-1], [-1], [1]]),
    )


def test_experiment_result_compat():
    er = ExperimentResult(setting=ExperimentSetting(plusX(0), sZ(0)), expectation=0.9, std_err=0.05, total_counts=100)
    assert str(er) == "X0_0→(1+0j)*Z0: 0.9 +- 0.05"


def test_experiment_result():
    er = ExperimentResult(setting=ExperimentSetting(plusX(0), sZ(0)), expectation=0.9, std_err=0.05, total_counts=100)
    assert str(er) == "X0_0→(1+0j)*Z0: 0.9 +- 0.05"


def test_ratio_variance_float():
    a, b, var_a, var_b = 1.0, 2.0, 0.1, 0.05
    ab_ratio_var = ratio_variance(a, var_a, b, var_b)
    assert ab_ratio_var == 0.028125


def test_ratio_variance_numerator_zero():
    # denominator can't be zero, but numerator can be
    a, b, var_a, var_b = 0.0, 2.0, 0.1, 0.05
    ab_ratio_var = ratio_variance(a, var_a, b, var_b)
    assert ab_ratio_var == 0.025


def test_ratio_variance_array():
    a = np.array([1.0, 10.0, 100.0])
    b = np.array([2.0, 20.0, 200.0])
    var_a = np.array([0.1, 1.0, 10.0])
    var_b = np.array([0.05, 0.5, 5.0])
    ab_ratio_var = ratio_variance(a, var_a, b, var_b)
    np.testing.assert_allclose(ab_ratio_var, np.array([0.028125, 0.0028125, 0.00028125]))


def test_correct_experiment_result():
    e = ExperimentResult(setting=ExperimentSetting(plusX(0), sZ(0)), expectation=0.9, std_err=0.05, total_counts=100)
    cal = ExperimentResult(setting=ExperimentSetting(plusX(0), sZ(0)), expectation=0.95, std_err=0.01, total_counts=100)
    corrected = ExperimentResult(
        setting=ExperimentSetting(plusX(0), sZ(0)),
        expectation=0.9473684210526316,
        std_err=np.sqrt(ratio_variance(0.9, 0.05 ** 2, 0.95, 0.01 ** 2)),
        raw_expectation=0.9,
        raw_std_err=0.05,
        calibration_expectation=0.95,
        calibration_std_err=0.01,
        calibration_counts=100,
        total_counts=100,
    )
    assert corrected == correct_experiment_result(e, cal)



#!/usr/bin/python
##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import math
import warnings
from functools import reduce
from itertools import product
from operator import mul

import numpy as np
import pytest

from pyquil.gates import RX, RZ, CNOT, H, X, PHASE
from pyquil.paulis import (
    PauliTerm,
    PauliSum,
    exponential_map,
    exponentiate_commuting_pauli_sum,
    ID,
    exponentiate,
    trotterize,
    is_zero,
    check_commutation,
    commuting_sets,
    term_with_coeff,
    sI,
    sX,
    sY,
    sZ,
    ZERO,
    is_identity,
)
from pyquil.quil import Program
from pyquil.simulation.tools import program_unitary


def isclose(a, b, rel_tol=1e-10, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def test_simplify_terms():
    term = PauliTerm("Z", 0) * -1.0 * PauliTerm("Z", 0)
    assert term.id() == ""
    assert term.coefficient == -1.0

    term = PauliTerm("Z", 0) + PauliTerm("Z", 0, 1.0)
    assert str(term) == "(2+0j)*Z0"


def test_get_qubits():
    term = PauliTerm("Z", 0) * PauliTerm("X", 1)
    assert term.get_qubits() == [0, 1]

    sum_term = PauliTerm("X", 0, 0.5) + 0.5j * PauliTerm("Y", 10) * PauliTerm("Y", 0, 0.5j)
    assert sum_term.get_qubits() == [0, 10]


def test_simplify_term_id_1():
    term = PauliTerm("I", 0, 0.5)
    assert term.id() == ""
    assert term.coefficient == 0.5


def test_simplify_term_id_2():
    term = 0.5 * ID()
    assert term.id() == ""
    assert term.coefficient == 0.5


def test_simplify_term_id_3():
    s = 0.25 + 0.25 * ID()
    terms = s.terms
    assert len(terms) == 1
    assert terms[0].id() == ""
    assert terms[0].coefficient == 0.5


def test_simplify_term_single():
    term = PauliTerm("Z", 0) * PauliTerm("I", 1) * PauliTerm("X", 2, 0.5j) * PauliTerm("Z", 0, 1.0)
    assert term.id() == "X2"
    assert term.coefficient == 0.5j


def test_simplify_term_xz():
    term1 = (-0.5 * PauliTerm("X", 0)) * (-1.0 * PauliTerm("Z", 0))
    term2 = -0.5 * PauliTerm("X", 0) * (-1.0) * PauliTerm("Z", 0)
    term3 = 0.5 * PauliTerm("X", 0) * PauliTerm("Z", 0)
    for term in [term1, term2, term3]:
        assert term.id() == "Y0"
        assert term.coefficient == -0.5j


def test_simplify_term_multindex():
    term = PauliTerm("X", 0, coefficient=-0.5) * PauliTerm("Z", 0, coefficient=-1.0) * PauliTerm("X", 2, 0.5)
    assert term.id(sort_ops=False) == "Y0X2"
    assert term.coefficient == -0.25j


def test_simplify_sum_terms():
    sum_term = PauliSum([PauliTerm("X", 0, 0.5), PauliTerm("Z", 0, 0.5j)])
    str_sum_term = str(sum_term + sum_term)
    assert str_sum_term == "(1+0j)*X0 + 1j*Z0" or str_sum_term == "1j*Z0 + (1+0j)*X0"
    sum_term = PauliSum([PauliTerm("X", 0, 0.5), PauliTerm("X", 0, 0.5)])
    assert str(sum_term.simplify()) == "(1+0j)*X0"

    # test the simplify on multiplication
    sum_term = PauliSum([PauliTerm("X", 0, 0.5), PauliTerm("X", 0, 0.5)])
    assert str(sum_term * sum_term) == "(1+0j)*I"


def test_copy():
    term = PauliTerm("X", 0, 0.5) * PauliTerm("X", 1, 0.5)
    new_term = term.copy()

    term = term * PauliTerm("X", 2, 0.5)
    new_term = new_term * PauliTerm("X", 2, 0.5)

    assert term == new_term  # value equality
    assert term is not new_term  # ref inequality
    assert term._ops is not new_term._ops

    term = PauliTerm("X", 0, 0.5) * PauliTerm("X", 1, 0.5)
    new_term = term * PauliTerm("X", 2, 0.5)
    assert term != new_term
    assert term is not new_term
    assert term._ops is not new_term._ops


def test_len():
    term = PauliTerm("Z", 0, 1.0) * PauliTerm("Z", 1, 1.0)
    assert len(term) == 2


def test_sum_len():
    pauli_sum = PauliTerm("Z", 0, 1.0) + PauliTerm("Z", 1, 1.0)
    assert len(pauli_sum) == 2


def test_enumerate():
    term = PauliTerm("Z", 0, 1.0) * PauliTerm("Z", 1, 1.0) * PauliTerm("X", 5, 5)
    position_op_pairs = [(0, "Z"), (1, "Z"), (5, "X")]
    for key, val in term:
        assert (key, val) in position_op_pairs


def test_getitem():
    term = PauliTerm("Z", 0, 1.0) * PauliTerm("Z", 1, 1.0) * PauliTerm("X", 5, 5)
    assert term[0] == "Z"
    assert term[1] == "Z"
    assert term[2] == "I"
    assert term[3] == "I"
    assert term[4] == "I"
    assert term[5] == "X"
    assert len(term) == 3


def test_ids():
    term_1 = PauliTerm("Z", 0, 1.0) * PauliTerm("Z", 1, 1.0) * PauliTerm("X", 5, 5)
    term_2 = PauliTerm("X", 5, 5) * PauliTerm("Z", 0, 1.0) * PauliTerm("Z", 1, 1.0)
    with pytest.warns(FutureWarning) as w:
        assert term_1.id() == term_2.id()
    assert "should be avoided" in str(w[0])


def test_ids_no_sort():
    term_1 = PauliTerm("Z", 0, 1.0) * PauliTerm("Z", 1, 1.0) * PauliTerm("X", 5, 5)
    term_2 = PauliTerm("X", 5, 5) * PauliTerm("Z", 0, 1.0) * PauliTerm("Z", 1, 1.0)
    assert term_1.id(sort_ops=False) == "Z0Z1X5"
    assert term_2.id(sort_ops=False) == "X5Z0Z1"


def test_operations_as_set():
    term_1 = PauliTerm("Z", 0, 1.0) * PauliTerm("Z", 1, 1.0) * PauliTerm("X", 5, 5)
    term_2 = PauliTerm("X", 5, 5) * PauliTerm("Z", 0, 1.0) * PauliTerm("Z", 1, 1.0)
    assert term_1.operations_as_set() == term_2.operations_as_set()


def test_pauliop_inputs():
    with pytest.raises(ValueError):
        PauliTerm("X", -2)


def test_pauli_sum():
    q_plus = 0.5 * PauliTerm("X", 0) + 0.5j * PauliTerm("Y", 0)
    the_sum = q_plus * PauliSum([PauliTerm("X", 0)])
    term_strings = [str(x) for x in the_sum.terms]
    assert "(0.5+0j)*I" in term_strings
    assert "(0.5+0j)*Z0" in term_strings
    assert len(term_strings) == 2
    assert len(the_sum.terms) == 2

    the_sum = q_plus * PauliTerm("X", 0)
    term_strings = [str(x) for x in the_sum.terms]
    assert "(0.5+0j)*I" in term_strings
    assert "(0.5+0j)*Z0" in term_strings
    assert len(term_strings) == 2
    assert len(the_sum.terms) == 2

    the_sum = PauliTerm("X", 0) * q_plus
    term_strings = [str(x) for x in the_sum.terms]
    assert "(0.5+0j)*I" in term_strings
    assert "(-0.5+0j)*Z0" in term_strings
    assert len(term_strings) == 2
    assert len(the_sum.terms) == 2

    with pytest.raises(ValueError):
        _ = PauliSum(sI(0))
    with pytest.raises(ValueError):
        _ = PauliSum([1, 1, 1, 1])
    with pytest.raises(ValueError):
        _ = the_sum * []


def test_ps_adds_pt_1():
    term = ID()
    b = term + term
    assert str(b) == "(2+0j)*I"
    assert str(b + term) == "(3+0j)*I"
    assert str(term + b) == "(3+0j)*I"


def test_ps_adds_pt_2():
    term = ID()
    b = term + 1.0
    assert str(b) == "(2+0j)*I"
    assert str(b + 1.0) == "(3+0j)*I"
    assert str(1.0 + b) == "(3+0j)*I"
    b = sX(0) + 1.0
    assert str(b) == "(1+0j)*X0 + (1+0j)*I"
    b = 1.0 + sX(0)
    assert str(b) == "(1+0j)*I + (1+0j)*X0"


def test_pauliterm_sub():
    assert str(sX(1) - 2.0) == str(sX(1) + -2.0)
    assert str(1.4 - sZ(1)) == str(1.4 + -1.0 * sZ(1))


def test_ps_sub():
    term = 3 * ID()
    b = term - 1.0
    assert str(b) == "(2+0j)*I"
    assert str(b - 1.0) == "(1+0j)*I"
    assert str(1.0 - b) == "(-1+0j)*I"
    b = 1.0 - sX(0)
    assert str(b) == "(1+0j)*I + (-1+0j)*X0"
    b = sX(0) - 1.0
    assert str(b) == "(1+0j)*X0 + (-1+0j)*I"


def test_zero_terms():
    term = PauliTerm("X", 0, 1.0) + PauliTerm("X", 0, -1.0) + PauliTerm("Y", 0, 0.5)
    assert str(term) == "(0.5+0j)*Y0"

    term = PauliTerm("X", 0, 1.0) + PauliTerm("X", 0, -1.0)
    assert str(term) == "0j*I"
    assert len(term.terms) == 1

    term2 = term * PauliTerm("Z", 2, 0.5)
    assert str(term2) == "0j*I"

    term3 = PauliTerm("Z", 2, 0.5) + term
    assert str(term3) == "(0.5+0j)*Z2"

    term4 = PauliSum([])
    assert str(term4) == "0j*I"

    term = PauliSum([PauliTerm("X", 0, 0.0), PauliTerm("Y", 1, 1.0) * PauliTerm("Z", 2)])
    assert str(term) == "0j*X0 + (1+0j)*Y1*Z2"
    term = term.simplify()
    assert str(term) == "(1+0j)*Y1*Z2"


def test_exponentiate_1():
    # test rotation of single qubit
    generator = PauliTerm("Z", 0, 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst(RZ(2.0, 0))
    assert prog == result_prog


def test_exponentiate_2():
    # testing general 2-circuit
    generator = PauliTerm("Z", 0, 1.0) * PauliTerm("Z", 1, 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst(CNOT(0, 1)).inst(RZ(2.0, 1)).inst(CNOT(0, 1))
    assert prog == result_prog


def test_exponentiate_bp0_ZX():
    # testing change of basis position 0
    generator = PauliTerm("X", 0, 1.0) * PauliTerm("Z", 1, 1.0)
    param_prog = exponential_map(generator)
    prog = param_prog(1)
    result_prog = Program().inst([H(0), CNOT(0, 1), RZ(2.0, 1), CNOT(0, 1), H(0)])
    assert prog == result_prog


def test_exponentiate_bp1_XZ():
    # testing change of basis position 1
    generator = PauliTerm("Z", 0, 1.0) * PauliTerm("X", 1, 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst([H(1), CNOT(0, 1), RZ(2.0, 1), CNOT(0, 1), H(1)])
    assert prog == result_prog


def test_exponentiate_bp0_ZY():
    # testing change of basis position 0
    generator = PauliTerm("Y", 0, 1.0) * PauliTerm("Z", 1, 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst([RX(math.pi / 2.0, 0), CNOT(0, 1), RZ(2.0, qubit=1), CNOT(0, 1), RX(-math.pi / 2, 0)])
    assert prog == result_prog


def test_exponentiate_bp1_YZ():
    # testing change of basis position 1
    generator = PauliTerm("Z", 0, 1.0) * PauliTerm("Y", 1, 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst([RX(math.pi / 2.0, 1), CNOT(0, 1), RZ(2.0, 1), CNOT(0, 1), RX(-math.pi / 2.0, 1)])
    assert prog == result_prog


def test_exponentiate_3cob():
    # testing circuit for 3-terms with change of basis
    generator = PauliTerm("Z", 0, 1.0) * PauliTerm("Y", 1, 1.0) * PauliTerm("X", 2, 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst(
        [
            RX(math.pi / 2.0, 1),
            H(2),
            CNOT(0, 1),
            CNOT(1, 2),
            RZ(2.0, 2),
            CNOT(1, 2),
            CNOT(0, 1),
            RX(-math.pi / 2.0, 1),
            H(2),
        ]
    )
    assert prog == result_prog


def test_exponentiate_3ns():
    # testing circuit for 3-terms non-sequential
    generator = PauliTerm("Y", 0, 1.0) * PauliTerm("I", 1, 1.0) * PauliTerm("Y", 2, 1.0) * PauliTerm("Y", 3, 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst(
        [
            RX(math.pi / 2.0, 0),
            RX(math.pi / 2.0, 2),
            RX(math.pi / 2.0, 3),
            CNOT(0, 2),
            CNOT(2, 3),
            RZ(2.0, 3),
            CNOT(2, 3),
            CNOT(0, 2),
            RX(-math.pi / 2.0, 0),
            RX(-math.pi / 2.0, 2),
            RX(-math.pi / 2.0, 3),
        ]
    )
    assert prog == result_prog


def test_exponentiate_commuting_pauli_sum():
    pauli_sum = PauliSum([PauliTerm("Z", 0, 0.5), PauliTerm("Z", 1, 0.5)])
    prog = Program().inst(RZ(1.0, 0)).inst(RZ(1.0, 1))
    result_prog = exponentiate_commuting_pauli_sum(pauli_sum)(1.0)
    assert prog == result_prog


def test_exponentiate_prog():
    ham = PauliTerm("Z", 0)
    result_prog = Program(RZ(2.0, 0))
    prog = exponentiate(ham)
    assert prog == result_prog


def test_exponentiate_identity():
    generator = PauliTerm("I", 1, 0.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program()
    assert prog == result_prog

    generator = PauliTerm("I", 1, 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst([X(0), PHASE(-1.0, 0), X(0), PHASE(-1.0, 0)])
    assert prog == result_prog

    generator = PauliTerm("I", 10, 0.08)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst([X(0), PHASE(-0.08, 0), X(0), PHASE(-0.08, 0)])
    assert prog == result_prog


def test_trotterize():
    term_one = PauliTerm("X", 0, 1.0)
    term_two = PauliTerm("Z", 0, 1.0)

    with pytest.raises(ValueError):
        trotterize(term_one, term_two, trotter_order=0)
    with pytest.raises(ValueError):
        trotterize(term_one, term_two, trotter_order=5)

    prog = trotterize(term_one, term_one)
    result_prog = Program().inst([H(0), RZ(2.0, 0), H(0), H(0), RZ(2.0, 0), H(0)])
    assert prog == result_prog

    # trotter_order 1 steps 1
    prog = trotterize(term_one, term_two, trotter_steps=1)
    result_prog = Program().inst([H(0), RZ(2.0, 0), H(0), RZ(2.0, 0)])
    assert prog == result_prog

    # trotter_order 1 steps 2
    prog = trotterize(term_one, term_two, trotter_steps=2)
    result_prog = Program().inst([H(0), RZ(1.0, 0), H(0), RZ(1.0, 0), H(0), RZ(1.0, 0), H(0), RZ(1.0, 0)])
    assert prog == result_prog

    # trotter_order 2 steps 1
    prog = trotterize(term_one, term_two, trotter_order=2)
    result_prog = Program().inst([H(0), RZ(1.0, 0), H(0), RZ(2.0, 0), H(0), RZ(1.0, 0), H(0)])
    assert prog == result_prog

    # trotter_order 2 steps 2
    prog = trotterize(term_one, term_two, trotter_order=2, trotter_steps=2)
    result_prog = Program().inst(
        [
            H(0),
            RZ(0.5, 0),
            H(0),
            RZ(1.0, 0),
            H(0),
            RZ(0.5, 0),
            H(0),
            H(0),
            RZ(0.5, 0),
            H(0),
            RZ(1.0, 0),
            H(0),
            RZ(0.5, 0),
            H(0),
        ]
    )
    assert prog == result_prog

    # trotter_order 3 steps 1
    prog = trotterize(term_one, term_two, trotter_order=3, trotter_steps=1)
    result_prog = Program().inst(
        [
            H(0),
            RZ(14.0 / 24, 0),
            H(0),
            RZ(4.0 / 3.0, 0),
            H(0),
            RZ(1.5, 0),
            H(0),
            RZ(-4.0 / 3.0, 0),
            H(0),
            RZ(-2.0 / 24, 0),
            H(0),
            RZ(2.0, 0),
        ]
    )
    assert prog == result_prog


def test_trotterize_order():
    def expmi(hermitian_matrix):
        """Compute the matrix exponential of -1j * hermitian_matrix."""
        L, Q = np.linalg.eigh(hermitian_matrix)
        return Q @ np.diag(np.exp(-1j * L)) @ Q.conj().T

    def error(order, time_step_length):
        a_pauli = time_step_length * sZ(0) * sY(1) * sX(2)
        a_program = a_pauli.program

        b_pauli = time_step_length * sX(0) * sZ(1) * sY(2)
        b_program = b_pauli.program

        num_qubits = len(a_program.get_qubits())
        assert num_qubits == len(b_program.get_qubits())

        a = program_unitary(a_program, num_qubits)
        b = program_unitary(b_program, num_qubits)
        a_plus_b = a + b
        exp_a_plus_b = expmi(time_step_length * a_plus_b)

        trotter_program = trotterize(a_pauli, b_pauli, trotter_order=order)
        trotter = program_unitary(trotter_program, num_qubits)

        return np.linalg.norm(exp_a_plus_b - trotter, np.inf)

    xs = 10 ** np.logspace(-1, -6, 10)
    for order in [1, 2, 3, 4]:
        ys = [error(order, float(x)) for x in xs]
        p = np.polyfit(np.log10(xs), np.log10(ys), 1)
        assert p[0] >= order, f"Bound not satisfied with order={order}: the slope is {p[0]}"


def test_is_zero():
    with pytest.raises(TypeError):
        is_zero(1)

    p_term = PauliTerm("X", 0)
    ps_term = p_term + PauliTerm("Z", 1)
    id_term = PauliTerm("I", 0)

    assert not is_zero(p_term)
    assert is_zero(p_term + -1 * p_term)
    assert not is_zero(ps_term)
    assert not is_zero(id_term)


def test_check_commutation():
    term1 = PauliTerm("X", 0) * PauliTerm("X", 1)
    term2 = PauliTerm("Y", 0) * PauliTerm("Y", 1)
    term3 = PauliTerm("Y", 0) * PauliTerm("Z", 2)
    # assert check_commutation(PauliSum([term1]), term2)
    assert check_commutation([term2], term3)
    assert check_commutation([term2], term3)
    assert not check_commutation([term1], term3)


def _commutator(t1, t2):
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=r"The term .+ will be combined with .+, " r"but they have different orders of operations.*",
            category=UserWarning,
        )
        return t1 * t2 + -1 * t2 * t1


def test_check_commutation_rigorous():
    # more rigorous test.  Get all operators in Pauli group
    p_n_group = ("I", "X", "Y", "Z")
    pauli_list = list(product(p_n_group, repeat=3))
    pauli_ops = [list(zip(x, range(3))) for x in pauli_list]
    pauli_ops_pq = [reduce(mul, (PauliTerm(*x) for x in op)) for op in pauli_ops]

    non_commuting_pairs = []
    commuting_pairs = []
    for x in range(len(pauli_ops_pq)):
        for y in range(x, len(pauli_ops_pq)):

            tmp_op = _commutator(pauli_ops_pq[x], pauli_ops_pq[y])
            assert len(tmp_op.terms) == 1
            if is_zero(tmp_op.terms[0]):
                commuting_pairs.append((pauli_ops_pq[x], pauli_ops_pq[y]))
            else:
                non_commuting_pairs.append((pauli_ops_pq[x], pauli_ops_pq[y]))

    # now that we have our sets let's check against our code.
    for t1, t2 in non_commuting_pairs:
        assert not check_commutation([t1], t2)

    for t1, t2 in commuting_pairs:
        assert check_commutation([t1], t2)


def test_commuting_sets():
    term1 = PauliTerm("X", 0) * PauliTerm("X", 1)
    term2 = PauliTerm("Y", 0) * PauliTerm("Y", 1)
    term3 = PauliTerm("Y", 0) * PauliTerm("Z", 2)
    pauli_sum = term1 + term2 + term3
    commuting_sets(pauli_sum)


def test_paulisum_iteration():
    term_list = [sX(2), sZ(4)]
    pauli_sum = sum(term_list)
    for ii, term in enumerate(pauli_sum):
        assert term_list[ii] == term


def test_paulisum_indexing():
    pauli_sum = 0.5 * sX(0) + 0.1 * sZ(1)
    assert pauli_sum[0] == 0.5 * sX(0)
    for ii, term in enumerate(pauli_sum.terms):
        assert pauli_sum[ii] == term


def test_term_powers():
    for qubit_id in range(2):
        pauli_terms = [sI(qubit_id), sX(qubit_id), sY(qubit_id), sZ(qubit_id)]
        for pauli_term in pauli_terms:
            assert pauli_term ** 0 == sI(qubit_id)
            assert pauli_term ** 1 == pauli_term
            assert pauli_term ** 2 == sI(qubit_id)
            assert pauli_term ** 3 == pauli_term
    with pytest.raises(ValueError):
        pauli_terms[0] ** -1
    # Test to make sure large powers can be computed
    (PauliTerm("X", 0, 2) * PauliTerm("Y", 0, 2)) ** 400


def test_sum_power():
    pauli_sum = (sY(0) - sX(0)) * (1.0 / np.sqrt(2))
    assert pauli_sum ** 2 == PauliSum([sI(0)])
    with pytest.raises(ValueError):
        _ = pauli_sum ** -1
    pauli_sum = sI(0) + sI(1)
    assert pauli_sum ** 0 == sI(0)
    # Test to make sure large powers can be computed
    pauli_sum ** 400


def test_term_equality():
    with pytest.raises(TypeError):
        sI(0) != 0
    assert sI(0) == sI(0)
    assert PauliTerm("X", 10, 1 + 1.0j) == PauliTerm("X", 10, 1 + 1.0j)
    assert PauliTerm("X", 10, 1 + 1.0j) + PauliTerm("X", 10, 1 + 1.0j) != PauliTerm("X", 10, 1 + 1.0j)
    assert PauliTerm("X", 10, 1 + 1.0j) != PauliTerm("X", 10, 1 + 1.0j) + PauliTerm("X", 10, 1 + 1.0j)


def test_term_with_coeff():
    assert PauliTerm("X", 0, 1.0j) == term_with_coeff(sX(0), 1.0j)
    assert PauliTerm("X", 0, -1.0) == term_with_coeff(sX(0), -1)
    with pytest.raises(ValueError):
        term_with_coeff(sI(0), None)


def test_sum_equality():
    pauli_sum = sY(0) - sX(0)
    assert pauli_sum != 2 * pauli_sum
    assert pauli_sum != pauli_sum + sZ(0)
    assert pauli_sum + sZ(0) != pauli_sum
    assert pauli_sum != sY(1) - sX(1)
    assert pauli_sum == -1.0 * sX(0) + sY(0)
    assert pauli_sum == pauli_sum * 1.0
    with pytest.raises(TypeError):
        assert pauli_sum != 0


def test_zero_term():
    qubit_id = 0
    coefficient = 10
    ps = sI(qubit_id) + sX(qubit_id)
    assert coefficient * ZERO() == ZERO()
    assert ZERO() * coefficient == ZERO()
    assert ZERO() * ID() == ZERO()
    assert ZERO() + ID() == ID()
    assert ZERO() + ps == ps
    assert ps + ZERO() == ps


def test_from_list():
    terms_list = [("X", 0), ("Y", 1), ("Z", 5)]
    term = reduce(lambda x, y: x * y, [PauliTerm(*x) for x in terms_list])

    pterm = PauliTerm.from_list(terms_list)
    assert pterm == term

    with pytest.raises(ValueError):
        # terms are not on disjoint qubits
        pterm = PauliTerm.from_list([("X", 0), ("Y", 0)])


def test_ordered():
    term = sZ(3) * sZ(2) * sZ(1)
    prog = exponential_map(term)(0.5)
    assert prog.out() == "CNOT 3 2\nCNOT 2 1\nRZ(1.0) 1\nCNOT 2 1\nCNOT 3 2\n"


def test_numpy_integer_types():
    (idx_np,) = np.arange(1, dtype=np.int64)
    assert isinstance(idx_np, np.int64)
    # on python 3 this fails unless explicitly allowing for numpy integer types
    PauliTerm("X", idx_np)


def test_simplify():
    t1 = sZ(0) * sZ(1)
    t2 = sZ(0) * sZ(1)
    assert (t1 + t2) == 2 * sZ(0) * sZ(1)


def test_dont_simplify():
    t1 = sZ(0) * sZ(1)
    t2 = sZ(2) * sZ(3)
    assert (t1 + t2) != 2 * sZ(0) * sZ(1)


def test_simplify_warning():
    t1 = sZ(0) * sZ(1)
    t2 = sZ(1) * sZ(0)
    tsum = t1 + t2
    assert tsum == 2 * sZ(0) * sZ(1)


def test_pauli_string():
    p = PauliTerm("X", 1) * PauliTerm("Z", 5)
    assert p.pauli_string([1, 5]) == "XZ"
    assert p.pauli_string([1]) == "X"
    assert p.pauli_string([5]) == "Z"
    assert p.pauli_string([5, 6]) == "ZI"
    assert p.pauli_string([0, 1]) == "IX"


def test_str():
    term = 2.0 * sX(1) * sX(2)
    assert str(term) == "(2+0j)*X1*X2"
    assert term.compact_str() == "(2+0j)*X1X2"


def test_from_str():
    with pytest.raises(ValueError):
        PauliTerm.from_compact_str("1*A0→1*Z0")


def test_is_identity():
    pt1 = -1.5j * sI(2)
    pt2 = 1.5 * sX(1) * sZ(2)

    assert is_identity(pt1)
    assert is_identity(pt2 + (-1 * pt2) + sI(0))
    assert not is_identity(0 * pt1)
    assert not is_identity(pt2 + (-1 * pt2))


def test_identity_no_qubit():
    assert is_identity(sI())


def test_qubit_validation():
    with pytest.raises(ValueError):
        sX(None)


def test_pauli_term_from_str():
    # tests that should _not_ fail are in test_pauli_sum_from_str
    with pytest.raises(ValueError):
        PauliTerm.from_compact_str("X0")
    with pytest.raises(ValueError):
        PauliTerm.from_compact_str("10")
    with pytest.raises(ValueError):
        PauliTerm.from_compact_str("1.0X0")
    with pytest.raises(ValueError):
        PauliTerm.from_compact_str("(1.0+9i)*X0")
    with pytest.raises(ValueError):
        PauliTerm.from_compact_str("(1.0+0j)*A0")


def test_pauli_sum_from_str():
    # this also tests PauliTerm.from_compact_str() since it gets called
    Sum = (1.5 + 0.5j) * sX(0) * sZ(2) + 0.7 * sZ(1)
    another_str = "(1.5 + 0.5j)*X0*Z2+.7*Z1"
    assert PauliSum.from_compact_str(str(Sum)) == Sum
    assert PauliSum.from_compact_str(Sum.compact_str()) == Sum
    assert PauliSum.from_compact_str(another_str) == Sum

    # test sums of length one
    Sum = PauliSum([1 * sY(0) * sY(1)])
    the_str = "1*Y0*Y1"
    assert PauliSum.from_compact_str(the_str) == Sum

    # test sums containing the identity
    Sum = (1.5 + 0.5j) * sX(0) * sZ(2) + 0.7 * sI(1)
    the_str = "(1.5 + 0.5j)*X0*Z2+.7*I"
    assert PauliSum.from_compact_str(the_str) == Sum

    # test the simplification (both in sums and products)
    Sum = PauliSum([2 * sY(1)])
    the_str = "1*Y0*X0 + (0+1j)*Z0 + 2*Y1"
    assert PauliSum.from_compact_str(the_str) == Sum



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import math
import re
import warnings
from functools import reduce
from itertools import product
from operator import mul

import numpy as np
import pytest

from pyquil.gates import RX, RZ, CNOT, H, X, PHASE
from pyquil.paulis import (
    PauliTerm,
    PauliSum,
    exponential_map,
    exponentiate_commuting_pauli_sum,
    ID,
    exponentiate,
    trotterize,
    is_zero,
    check_commutation,
    commuting_sets,
    term_with_coeff,
    sI,
    sX,
    sY,
    sZ,
    ZERO,
)
from pyquil.quil import Program, address_qubits, get_default_qubit_mapping
from pyquil.quilatom import QubitPlaceholder


def test_simplify_terms():
    q = QubitPlaceholder.register(1)
    term = PauliTerm("Z", q[0]) * -1.0 * PauliTerm("Z", q[0])
    assert term.id() == ""
    assert term.coefficient == -1.0

    term = PauliTerm("Z", q[0]) + PauliTerm("Z", q[0], 1.0)
    assert str(term).startswith("(2+0j)*Zq")


def test_get_qubits():
    q = QubitPlaceholder.register(2)
    term = PauliTerm("Z", q[0]) * PauliTerm("X", q[1])
    assert term.get_qubits() == q

    q10 = QubitPlaceholder()
    sum_term = PauliTerm("X", q[0], 0.5) + 0.5j * PauliTerm("Y", q10) * PauliTerm("Y", q[0], 0.5j)
    assert sum_term.get_qubits() == [q[0], q10]


def test_simplify_term_single():
    q0, q1, q2 = QubitPlaceholder.register(3)
    term = PauliTerm("Z", q0) * PauliTerm("I", q1) * PauliTerm("X", q2, 0.5j) * PauliTerm("Z", q0, 1.0)
    assert term.id() == "X{}".format(q2)
    assert term.coefficient == 0.5j


def test_simplify_term_xz():
    q0 = QubitPlaceholder()
    term1 = (-0.5 * PauliTerm("X", q0)) * (-1.0 * PauliTerm("Z", q0))
    term2 = -0.5 * PauliTerm("X", q0) * (-1.0) * PauliTerm("Z", q0)
    term3 = 0.5 * PauliTerm("X", q0) * PauliTerm("Z", q0)
    for term in [term1, term2, term3]:
        assert term.id() == "Y{}".format(q0)
        assert term.coefficient == -0.5j


def test_simplify_term_multindex():
    q0, q2 = QubitPlaceholder.register(2)
    term = PauliTerm("X", q0, coefficient=-0.5) * PauliTerm("Z", q0, coefficient=-1.0) * PauliTerm("X", q2, 0.5)
    assert term.id(sort_ops=False) == "Y{q0}X{q2}".format(q0=q0, q2=q2)
    assert term.coefficient == -0.25j


def test_simplify_sum_terms():
    q0 = QubitPlaceholder()
    sum_term = PauliSum([PauliTerm("X", q0, 0.5), PauliTerm("Z", q0, 0.5j)])
    str_sum_term = str(sum_term + sum_term)
    assert str_sum_term == "(1+0j)*X{q0} + 1j*Z{q0}".format(q0=q0) or str_sum_term == "1j*Z{q0} + (1+0j)*X{q0}".format(
        q0=q0
    )
    sum_term = PauliSum([PauliTerm("X", q0, 0.5), PauliTerm("X", q0, 0.5)])
    assert str(sum_term.simplify()) == "(1+0j)*X{q0}".format(q0=q0)

    # test the simplify on multiplication
    sum_term = PauliSum([PauliTerm("X", q0, 0.5), PauliTerm("X", q0, 0.5)])
    assert str(sum_term * sum_term) == "(1+0j)*I"


def test_copy():
    q0, q1 = QubitPlaceholder.register(2)
    term = PauliTerm("X", q0, 0.5) * PauliTerm("X", q1, 0.5)
    new_term = term.copy()

    q2 = QubitPlaceholder()
    term = term * PauliTerm("X", q2, 0.5)
    new_term = new_term * PauliTerm("X", q2, 0.5)

    assert term == new_term  # value equality
    assert term is not new_term  # ref inequality
    assert term._ops is not new_term._ops

    term = PauliTerm("X", q0, 0.5) * PauliTerm("X", q1, 0.5)
    new_term = term * PauliTerm("X", q2, 0.5)
    assert term != new_term
    assert term is not new_term
    assert term._ops is not new_term._ops


def test_len():
    q0, q1 = QubitPlaceholder.register(2)
    term = PauliTerm("Z", q0, 1.0) * PauliTerm("Z", q1, 1.0)
    assert len(term) == 2


def test_sum_len():
    q0, q1 = QubitPlaceholder.register(2)
    pauli_sum = PauliTerm("Z", q0, 1.0) + PauliTerm("Z", q1, 1.0)
    assert len(pauli_sum) == 2


def test_enumerate():
    q0, q1, q5 = QubitPlaceholder.register(3)
    term = PauliTerm("Z", q0, 1.0) * PauliTerm("Z", q1, 1.0) * PauliTerm("X", q5, 5)
    position_op_pairs = [(q0, "Z"), (q1, "Z"), (q5, "X")]
    for key, val in term:
        assert (key, val) in position_op_pairs


def test_getitem():
    q = QubitPlaceholder.register(6)
    term = PauliTerm("Z", q[0], 1.0) * PauliTerm("Z", q[1], 1.0) * PauliTerm("X", q[5], 5)
    assert term[q[0]] == "Z"
    assert term[q[1]] == "Z"
    assert term[q[2]] == "I"
    assert term[q[3]] == "I"
    assert term[q[4]] == "I"
    assert term[q[5]] == "X"
    assert len(term) == 3


def test_ids():
    q = QubitPlaceholder.register(6)
    term_1 = PauliTerm("Z", q[0], 1.0) * PauliTerm("Z", q[1], 1.0) * PauliTerm("X", q[5], 5)
    term_2 = PauliTerm("X", q[5], 5) * PauliTerm("Z", q[0], 1.0) * PauliTerm("Z", q[1], 1.0)
    # Not sortable
    with pytest.raises(TypeError):
        with pytest.warns(FutureWarning):
            term_1.id() == term_2.id()


def test_ids_no_sort():
    q = QubitPlaceholder.register(6)
    term_1 = PauliTerm("Z", q[0], 1.0) * PauliTerm("Z", q[1], 1.0) * PauliTerm("X", q[5], 5)
    term_2 = PauliTerm("X", q[5], 5) * PauliTerm("Z", q[0], 1.0) * PauliTerm("Z", q[1], 1.0)
    assert re.match("Z.+Z.+X.+", term_1.id(sort_ops=False))
    assert re.match("X.+Z.+Z.+", term_2.id(sort_ops=False))


def test_operations_as_set():
    q = QubitPlaceholder.register(6)
    term_1 = PauliTerm("Z", q[0], 1.0) * PauliTerm("Z", q[1], 1.0) * PauliTerm("X", q[5], 5)
    term_2 = PauliTerm("X", q[5], 5) * PauliTerm("Z", q[0], 1.0) * PauliTerm("Z", q[1], 1.0)
    assert term_1.operations_as_set() == term_2.operations_as_set()


def test_pauli_sum():
    q = QubitPlaceholder.register(8)
    q_plus = 0.5 * PauliTerm("X", q[0]) + 0.5j * PauliTerm("Y", q[0])
    the_sum = q_plus * PauliSum([PauliTerm("X", q[0])])
    term_strings = [str(x) for x in the_sum.terms]
    assert "(0.5+0j)*I" in term_strings
    assert len(term_strings) == 2
    assert len(the_sum.terms) == 2

    the_sum = q_plus * PauliTerm("X", q[0])
    term_strings = [str(x) for x in the_sum.terms]
    assert "(0.5+0j)*I" in term_strings
    assert len(term_strings) == 2
    assert len(the_sum.terms) == 2

    the_sum = PauliTerm("X", q[0]) * q_plus
    term_strings = [str(x) for x in the_sum.terms]
    assert "(0.5+0j)*I" in term_strings
    assert len(term_strings) == 2
    assert len(the_sum.terms) == 2

    with pytest.raises(ValueError):
        _ = PauliSum(sI(q[0]))
    with pytest.raises(ValueError):
        _ = PauliSum([1, 1, 1, 1])
    with pytest.raises(ValueError):
        _ = the_sum * []


def test_pauliterm_sub():
    q = QubitPlaceholder.register(8)
    assert str(sX(q[1]) - 2.0) == str(sX(q[1]) + -2.0)
    assert str(1.4 - sZ(q[1])) == str(1.4 + -1.0 * sZ(q[1]))


def test_ps_sub():
    q0 = QubitPlaceholder()
    term = 3 * ID()
    b = term - 1.0
    assert str(b) == "(2+0j)*I"
    assert str(b - 1.0) == "(1+0j)*I"
    assert str(1.0 - b) == "(-1+0j)*I"
    b = 1.0 - sX(q0)
    assert re.match(r"\(1\+0j\)\*I \+ \(-1\+0j\)\*Xq\d+", str(b))
    b = sX(q0) - 1.0
    assert re.match(r"\(1\+0j\)\*Xq\d+ \+ \(-1\+0j\)\*I", str(b))


def test_exponentiate_1():
    # test rotation of single qubit
    q = QubitPlaceholder.register(8)
    generator = PauliTerm("Z", q[0], 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst(RZ(2.0, q[0]))
    assert address_qubits(prog) == address_qubits(result_prog)


def test_exponentiate_2():
    # testing general 2-circuit
    q = QubitPlaceholder.register(8)
    generator = PauliTerm("Z", q[0], 1.0) * PauliTerm("Z", q[1], 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst(CNOT(q[0], q[1])).inst(RZ(2.0, q[1])).inst(CNOT(q[0], q[1]))

    mapping = get_default_qubit_mapping(prog)
    assert address_qubits(prog, mapping) == address_qubits(result_prog, mapping)


def test_exponentiate_bp0_ZX():
    q = QubitPlaceholder.register(8)
    # testing change of basis position 0
    generator = PauliTerm("X", q[0], 1.0) * PauliTerm("Z", q[1], 1.0)
    param_prog = exponential_map(generator)
    prog = param_prog(1)
    result_prog = Program().inst([H(q[0]), CNOT(q[0], q[1]), RZ(2.0, q[1]), CNOT(q[0], q[1]), H(q[0])])
    assert address_qubits(prog) == address_qubits(result_prog)


def test_exponentiate_bp1_XZ():
    # testing change of basis position 1
    q = QubitPlaceholder.register(8)
    generator = PauliTerm("Z", q[0], 1.0) * PauliTerm("X", q[1], 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst([H(q[1]), CNOT(q[0], q[1]), RZ(2.0, q[1]), CNOT(q[0], q[1]), H(q[1])])
    assert address_qubits(prog) == address_qubits(result_prog)


def test_exponentiate_bp0_ZY():
    # testing change of basis position 0
    q = QubitPlaceholder.register(8)
    generator = PauliTerm("Y", q[0], 1.0) * PauliTerm("Z", q[1], 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst(
        [
            RX(math.pi / 2.0, q[0]),
            CNOT(q[0], q[1]),
            RZ(2.0, q[1]),
            CNOT(q[0], q[1]),
            RX(-math.pi / 2, q[0]),
        ]
    )
    assert address_qubits(prog) == address_qubits(result_prog)


def test_exponentiate_bp1_YZ():
    q = QubitPlaceholder.register(8)
    # testing change of basis position 1
    generator = PauliTerm("Z", q[0], 1.0) * PauliTerm("Y", q[1], 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst(
        [
            RX(math.pi / 2.0, q[1]),
            CNOT(q[0], q[1]),
            RZ(2.0, q[1]),
            CNOT(q[0], q[1]),
            RX(-math.pi / 2.0, q[1]),
        ]
    )
    assert address_qubits(prog) == address_qubits(result_prog)


def test_exponentiate_3cob():
    # testing circuit for 3-terms with change of basis
    q = QubitPlaceholder.register(8)
    generator = PauliTerm("Z", q[0], 1.0) * PauliTerm("Y", q[1], 1.0) * PauliTerm("X", q[2], 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst(
        [
            RX(math.pi / 2.0, q[1]),
            H(q[2]),
            CNOT(q[0], q[1]),
            CNOT(q[1], q[2]),
            RZ(2.0, q[2]),
            CNOT(q[1], q[2]),
            CNOT(q[0], q[1]),
            RX(-math.pi / 2.0, q[1]),
            H(q[2]),
        ]
    )
    assert address_qubits(prog) == address_qubits(result_prog)


def test_exponentiate_3ns():
    # testing circuit for 3-terms non-sequential
    q = QubitPlaceholder.register(8)
    generator = (
        PauliTerm("Y", q[0], 1.0) * PauliTerm("I", q[1], 1.0) * PauliTerm("Y", q[2], 1.0) * PauliTerm("Y", q[3], 1.0)
    )
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst(
        [
            RX(math.pi / 2.0, q[0]),
            RX(math.pi / 2.0, q[2]),
            RX(math.pi / 2.0, q[3]),
            CNOT(q[0], q[2]),
            CNOT(q[2], q[3]),
            RZ(2.0, q[3]),
            CNOT(q[2], q[3]),
            CNOT(q[0], q[2]),
            RX(-math.pi / 2.0, q[0]),
            RX(-math.pi / 2.0, q[2]),
            RX(-math.pi / 2.0, q[3]),
        ]
    )
    assert address_qubits(prog) == address_qubits(result_prog)


def test_exponentiate_commuting_pauli_sum():
    q = QubitPlaceholder.register(8)
    pauli_sum = PauliSum([PauliTerm("Z", q[0], 0.5), PauliTerm("Z", q[1], 0.5)])
    prog = Program().inst(RZ(1.0, q[0])).inst(RZ(1.0, q[1]))
    result_prog = exponentiate_commuting_pauli_sum(pauli_sum)(1.0)
    assert address_qubits(prog) == address_qubits(result_prog)


def test_exponentiate_prog():
    q = QubitPlaceholder.register(8)
    ham = PauliTerm("Z", q[0])
    result_prog = Program(RZ(2.0, q[0]))
    prog = exponentiate(ham)
    assert address_qubits(prog) == address_qubits(result_prog)


def test_exponentiate_identity():
    q = QubitPlaceholder.register(11)

    generator = PauliTerm("I", q[1], 0.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    assert len(prog) == 0

    generator = PauliTerm("I", q[1], 1.0)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst([X(q[0]), PHASE(-1.0, q[0]), X(q[0]), PHASE(-1.0, q[0])])
    assert address_qubits(prog) == address_qubits(result_prog)

    generator = PauliTerm("I", q[10], 0.08)
    para_prog = exponential_map(generator)
    prog = para_prog(1)
    result_prog = Program().inst([X(q[0]), PHASE(-0.08, q[0]), X(q[0]), PHASE(-0.08, q[0])])
    assert address_qubits(prog) == address_qubits(result_prog)


def test_trotterize():
    q = QubitPlaceholder.register(8)
    term_one = PauliTerm("X", q[0], 1.0)
    term_two = PauliTerm("Z", q[0], 1.0)

    with pytest.raises(ValueError):
        trotterize(term_one, term_two, trotter_order=0)
    with pytest.raises(ValueError):
        trotterize(term_one, term_two, trotter_order=5)

    prog = trotterize(term_one, term_one)
    result_prog = Program().inst([H(q[0]), RZ(2.0, q[0]), H(q[0]), H(q[0]), RZ(2.0, q[0]), H(q[0])])
    assert address_qubits(prog) == address_qubits(result_prog)

    # trotter_order 1 steps 1
    prog = trotterize(term_one, term_two, trotter_steps=1)
    result_prog = Program().inst([H(q[0]), RZ(2.0, q[0]), H(q[0]), RZ(2.0, q[0])])
    assert address_qubits(prog) == address_qubits(result_prog)

    # trotter_order 1 steps 2
    prog = trotterize(term_one, term_two, trotter_steps=2)
    result_prog = Program().inst(
        [
            H(q[0]),
            RZ(1.0, q[0]),
            H(q[0]),
            RZ(1.0, q[0]),
            H(q[0]),
            RZ(1.0, q[0]),
            H(q[0]),
            RZ(1.0, q[0]),
        ]
    )
    assert address_qubits(prog) == address_qubits(result_prog)

    # trotter_order 2 steps 1
    prog = trotterize(term_one, term_two, trotter_order=2)
    result_prog = Program().inst([H(q[0]), RZ(1.0, q[0]), H(q[0]), RZ(2.0, q[0]), H(q[0]), RZ(1.0, q[0]), H(q[0])])
    assert address_qubits(prog) == address_qubits(result_prog)

    # trotter_order 2 steps 2
    prog = trotterize(term_one, term_two, trotter_order=2, trotter_steps=2)
    result_prog = Program().inst(
        [
            H(q[0]),
            RZ(0.5, q[0]),
            H(q[0]),
            RZ(1.0, q[0]),
            H(q[0]),
            RZ(0.5, q[0]),
            H(q[0]),
            H(q[0]),
            RZ(0.5, q[0]),
            H(q[0]),
            RZ(1.0, q[0]),
            H(q[0]),
            RZ(0.5, q[0]),
            H(q[0]),
        ]
    )
    assert address_qubits(prog) == address_qubits(result_prog)

    # trotter_order 3 steps 1
    prog = trotterize(term_one, term_two, trotter_order=3, trotter_steps=1)
    result_prog = Program().inst(
        [
            H(q[0]),
            RZ(14.0 / 24, q[0]),
            H(q[0]),
            RZ(4.0 / 3.0, q[0]),
            H(q[0]),
            RZ(1.5, q[0]),
            H(q[0]),
            RZ(-4.0 / 3.0, q[0]),
            H(q[0]),
            RZ(-2.0 / 24, q[0]),
            H(q[0]),
            RZ(2.0, q[0]),
        ]
    )
    assert address_qubits(prog) == address_qubits(result_prog)


def test_is_zeron():
    q = QubitPlaceholder.register(8)
    with pytest.raises(TypeError):
        is_zero(1)

    p_term = PauliTerm("X", q[0])
    ps_term = p_term + PauliTerm("Z", q[1])

    assert not is_zero(p_term)
    assert is_zero(p_term + -1 * p_term)
    assert not is_zero(ps_term)


def test_check_commutation():
    q = QubitPlaceholder.register(8)
    term1 = PauliTerm("X", q[0]) * PauliTerm("X", q[1])
    term2 = PauliTerm("Y", q[0]) * PauliTerm("Y", q[1])
    term3 = PauliTerm("Y", q[0]) * PauliTerm("Z", q[2])
    # assert check_commutation(PauliSum([term1]), term2)
    assert check_commutation([term2], term3)
    assert check_commutation([term2], term3)
    assert not check_commutation([term1], term3)


def _commutator(t1, t2):
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=r"The term .+ will be combined with .+, " r"but they have different orders of operations.*",
            category=UserWarning,
        )
        return t1 * t2 + -1 * t2 * t1


def test_check_commutation_rigorous():
    # more rigorous test.  Get all operators in Pauli group
    p_n_group = ("I", "X", "Y", "Z")
    pauli_list = list(product(p_n_group, repeat=3))
    pauli_ops = [list(zip(x, QubitPlaceholder.register(3))) for x in pauli_list]
    pauli_ops_pq = [reduce(mul, (PauliTerm(*x) for x in op)) for op in pauli_ops]

    non_commuting_pairs = []
    commuting_pairs = []
    for x in range(len(pauli_ops_pq)):
        for y in range(x, len(pauli_ops_pq)):

            tmp_op = _commutator(pauli_ops_pq[x], pauli_ops_pq[y])
            assert len(tmp_op.terms) == 1
            if is_zero(tmp_op.terms[0]):
                commuting_pairs.append((pauli_ops_pq[x], pauli_ops_pq[y]))
            else:
                non_commuting_pairs.append((pauli_ops_pq[x], pauli_ops_pq[y]))

    # now that we have our sets let's check against our code.
    for t1, t2 in non_commuting_pairs:
        assert not check_commutation([t1], t2)

    for t1, t2 in commuting_pairs:
        assert check_commutation([t1], t2)


def test_commuting_sets():
    q = QubitPlaceholder.register(8)
    term1 = PauliTerm("X", q[0]) * PauliTerm("X", q[1])
    term2 = PauliTerm("Y", q[0]) * PauliTerm("Y", q[1])
    term3 = PauliTerm("Y", q[0]) * PauliTerm("Z", q[2])
    pauli_sum = term1 + term2 + term3
    commuting_sets(pauli_sum)


def test_paulisum_iteration():
    q = QubitPlaceholder.register(8)
    term_list = [sX(q[2]), sZ(q[4])]
    pauli_sum = sum(term_list)
    for ii, term in enumerate(pauli_sum):
        assert term_list[ii] == term


def test_paulisum_indexing():
    q = QubitPlaceholder.register(8)
    pauli_sum = 0.5 * sX(q[0]) + 0.1 * sZ(q[1])
    assert pauli_sum[0] == 0.5 * sX(q[0])
    for ii, term in enumerate(pauli_sum.terms):
        assert pauli_sum[ii] == term


def test_term_powers():
    for qubit in QubitPlaceholder.register(2):
        pauli_terms = [sI(qubit), sX(qubit), sY(qubit), sZ(qubit)]
        for pauli_term in pauli_terms:
            assert pauli_term ** 0 == sI(qubit)
            assert pauli_term ** 1 == pauli_term
            assert pauli_term ** 2 == sI(qubit)
            assert pauli_term ** 3 == pauli_term
    with pytest.raises(ValueError):
        pauli_terms[0] ** -1


def test_term_large_powers():
    # Test to make sure large powers can be computed
    q = QubitPlaceholder.register(2)
    (PauliTerm("X", q[0], 2) * PauliTerm("Y", q[0], 2)) ** 400


def test_sum_power():
    q = QubitPlaceholder.register(8)
    pauli_sum = (sY(q[0]) - sX(q[0])) * (1.0 / np.sqrt(2))
    assert pauli_sum ** 2 == PauliSum([sI(q[0])])
    with pytest.raises(ValueError):
        _ = pauli_sum ** -1
    pauli_sum = sI(q[0]) + sI(q[1])
    assert pauli_sum ** 0 == sI(q[0])
    # Test to make sure large powers can be computed
    pauli_sum ** 400


def test_term_equality():
    q0, q10 = QubitPlaceholder.register(2)
    with pytest.raises(TypeError):
        sI(q0) != 0
    assert sI(q0) == sI(q0)
    assert PauliTerm("X", q10, 1 + 1.0j) == PauliTerm("X", q10, 1 + 1.0j)
    assert PauliTerm("X", q10, 1 + 1.0j) + PauliTerm("X", q10, 1 + 1.0j) != PauliTerm("X", q10, 1 + 1.0j)
    assert PauliTerm("X", q10, 1 + 1.0j) != PauliTerm("X", q10, 1 + 1.0j) + PauliTerm("X", q10, 1 + 1.0j)


def test_term_with_coeff():
    q0 = QubitPlaceholder()
    assert PauliTerm("X", q0, 1.0j) == term_with_coeff(sX(q0), 1.0j)
    assert PauliTerm("X", q0, -1.0) == term_with_coeff(sX(q0), -1)
    with pytest.raises(ValueError):
        term_with_coeff(sI(q0), None)


def test_sum_equality():
    q0, q1 = QubitPlaceholder.register(2)
    pauli_sum = sY(q0) - sX(q0)
    assert pauli_sum != 2 * pauli_sum
    assert pauli_sum != pauli_sum + sZ(q0)
    assert pauli_sum + sZ(q0) != pauli_sum
    assert pauli_sum != sY(q1) - sX(q1)
    assert pauli_sum == -1.0 * sX(q0) + sY(q0)
    assert pauli_sum == pauli_sum * 1.0
    with pytest.raises(TypeError):
        assert pauli_sum != 0


def test_zero_term():
    qubit_id = QubitPlaceholder()
    coefficient = 10
    ps = sI(qubit_id) + sX(qubit_id)
    assert coefficient * ZERO() == ZERO()
    assert ZERO() * coefficient == ZERO()
    assert ZERO() * ID() == ZERO()
    assert ZERO() + ID() == ID()
    assert ZERO() + ps == ps
    assert ps + ZERO() == ps


def test_from_list():
    q = QubitPlaceholder.register(8)
    terms_list = [("X", q[0]), ("Y", q[1]), ("Z", q[5])]
    term = reduce(lambda x, y: x * y, [PauliTerm(*x) for x in terms_list])

    pterm = PauliTerm.from_list(terms_list)
    assert pterm == term

    with pytest.raises(ValueError):
        # terms are not on disjoint qubits
        PauliTerm.from_list([("X", q[0]), ("Y", q[0])])


def test_ordered():
    q = QubitPlaceholder.register(8)
    mapping = {x: i for i, x in enumerate(q)}
    term = sZ(q[3]) * sZ(q[2]) * sZ(q[1])
    prog = address_qubits(exponential_map(term)(0.5), mapping)
    assert prog.out() == "CNOT 3 2\nCNOT 2 1\nRZ(1.0) 1\nCNOT 2 1\nCNOT 3 2\n"


def test_simplify():
    q = QubitPlaceholder.register(8)
    t1 = sZ(q[0]) * sZ(q[1])
    t2 = sZ(q[0]) * sZ(q[1])
    assert (t1 + t2) == 2 * sZ(q[0]) * sZ(q[1])


def test_dont_simplify():
    q = QubitPlaceholder.register(8)
    t1 = sZ(q[0]) * sZ(q[1])
    t2 = sZ(q[2]) * sZ(q[3])
    assert (t1 + t2) != 2 * sZ(q[0]) * sZ(q[1])


def test_simplify_warning():
    q = QubitPlaceholder.register(8)
    t1 = sZ(q[0]) * sZ(q[1])
    t2 = sZ(q[1]) * sZ(q[0])
    tsum = t1 + t2
    assert tsum == 2 * sZ(q[0]) * sZ(q[1])



import numpy as np

from pyquil.simulation.matrices import (
    QUANTUM_GATES,
    relaxation_operators,
    dephasing_operators,
    depolarizing_operators,
    bit_flip_operators,
    phase_flip_operators,
    bitphase_flip_operators,
)


def test_singleq():
    assert np.isclose(QUANTUM_GATES["I"], np.eye(2)).all()
    assert np.isclose(QUANTUM_GATES["X"], np.array([[0, 1], [1, 0]])).all()
    assert np.isclose(QUANTUM_GATES["Y"], np.array([[0, -1j], [1j, 0]])).all()
    assert np.isclose(QUANTUM_GATES["Z"], np.array([[1, 0], [0, -1]])).all()

    assert np.isclose(QUANTUM_GATES["H"], (1.0 / np.sqrt(2)) * np.array([[1, 1], [1, -1]])).all()
    assert np.isclose(QUANTUM_GATES["S"], np.array([[1.0, 0], [0, 1j]])).all()
    assert np.isclose(QUANTUM_GATES["T"], np.array([[1.0, 0.0], [0.0, np.exp(1.0j * np.pi / 4.0)]])).all()


def test_parametric():
    phi_range = np.linspace(0, 2 * np.pi, 120)
    for phi in phi_range:
        assert np.isclose(QUANTUM_GATES["PHASE"](phi), np.array([[1.0, 0.0], [0.0, np.exp(1j * phi)]])).all()
        assert np.isclose(
            QUANTUM_GATES["RX"](phi),
            np.array(
                [
                    [np.cos(phi / 2.0), -1j * np.sin(phi / 2.0)],
                    [-1j * np.sin(phi / 2.0), np.cos(phi / 2.0)],
                ]
            ),
        ).all()
        assert np.isclose(
            QUANTUM_GATES["RY"](phi),
            np.array([[np.cos(phi / 2.0), -np.sin(phi / 2.0)], [np.sin(phi / 2.0), np.cos(phi / 2.0)]]),
        ).all()
        assert np.isclose(
            QUANTUM_GATES["RZ"](phi),
            np.array(
                [
                    [np.cos(phi / 2.0) - 1j * np.sin(phi / 2.0), 0],
                    [0, np.cos(phi / 2.0) + 1j * np.sin(phi / 2.0)],
                ]
            ),
        ).all()


def test_multiq():
    assert np.isclose(QUANTUM_GATES["CNOT"], np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])).all()
    assert np.isclose(
        QUANTUM_GATES["CCNOT"],
        np.array(
            [
                [1, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 0, 1, 0],
            ]
        ),
    ).all()


def test_kraus_t1_normalization():
    kraus_ops = relaxation_operators(0.75)
    total = np.zeros((2, 2), dtype=complex)
    for kop in kraus_ops:
        total += np.conj(kop.T).dot(kop)
    assert np.allclose(total, np.eye(2))


def test_kraus_t2_normalization():
    kraus_ops = dephasing_operators(0.75)
    total = np.zeros((2, 2), dtype=complex)
    for kop in kraus_ops:
        total += np.conj(kop.T).dot(kop)
    assert np.allclose(total, np.eye(2))


def test_kraus_depolarizing_normalization():
    kraus_ops = depolarizing_operators(0.75)
    total = np.zeros((2, 2), dtype=complex)
    for kop in kraus_ops:
        total += np.conj(kop.T).dot(kop)
    assert np.allclose(total, np.eye(2))


def test_kraus_bitflip_normalization():
    kraus_ops = bit_flip_operators(0.75)
    total = np.zeros((2, 2), dtype=complex)
    for kop in kraus_ops:
        total += np.conj(kop.T).dot(kop)
    assert np.allclose(total, np.eye(2))


def test_kraus_phaseflip_normalization():
    kraus_ops = phase_flip_operators(0.75)
    total = np.zeros((2, 2), dtype=complex)
    for kop in kraus_ops:
        total += np.conj(kop.T).dot(kop)
    assert np.allclose(total, np.eye(2))


def test_kraus_bitphaseflip_normalization():
    kraus_ops = bitphase_flip_operators(0.75)
    total = np.zeros((2, 2), dtype=complex)
    for kop in kraus_ops:
        total += np.conj(kop.T).dot(kop)
    assert np.allclose(total, np.eye(2))



import numpy as np

from pyquil import Program
from pyquil.api import QVM, QuantumComputer, get_qc, Client
from pyquil.experiment import ExperimentSetting, Experiment
from pyquil.gates import CNOT, H, RESET, RY, X
from pyquil.noise import NoiseModel
from pyquil.paulis import sX, sY, sZ
from pyquil.tests.utils import DummyCompiler


def test_qc_expectation(client: Client, dummy_compiler: DummyCompiler):
    qc = QuantumComputer(name="testy!", qam=QVM(client=client), compiler=dummy_compiler)

    # bell state program
    p = Program()
    p += RESET()
    p += H(0)
    p += CNOT(0, 1)
    p.wrap_in_numshots_loop(10)

    # XX, YY, ZZ experiment
    sx = ExperimentSetting(in_state=sZ(0) * sZ(1), out_operator=sX(0) * sX(1))
    sy = ExperimentSetting(in_state=sZ(0) * sZ(1), out_operator=sY(0) * sY(1))
    sz = ExperimentSetting(in_state=sZ(0) * sZ(1), out_operator=sZ(0) * sZ(1))

    e = Experiment(settings=[sx, sy, sz], program=p)

    results = qc.experiment(e)

    # XX expectation value for bell state |00> + |11> is 1
    assert np.isclose(results[0].expectation, 1)
    assert np.isclose(results[0].std_err, 0)
    assert results[0].total_counts == 40

    # YY expectation value for bell state |00> + |11> is -1
    assert np.isclose(results[1].expectation, -1)
    assert np.isclose(results[1].std_err, 0)
    assert results[1].total_counts == 40

    # ZZ expectation value for bell state |00> + |11> is 1
    assert np.isclose(results[2].expectation, 1)
    assert np.isclose(results[2].std_err, 0)
    assert results[2].total_counts == 40


def test_qc_expectation_larger_lattice(client: Client, dummy_compiler: DummyCompiler):
    qc = QuantumComputer(name="testy!", qam=QVM(client=client), compiler=dummy_compiler)

    q0 = 2
    q1 = 3

    # bell state program
    p = Program()
    p += RESET()
    p += H(q0)
    p += CNOT(q0, q1)
    p.wrap_in_numshots_loop(10)

    # XX, YY, ZZ experiment
    sx = ExperimentSetting(in_state=sZ(q0) * sZ(q1), out_operator=sX(q0) * sX(q1))
    sy = ExperimentSetting(in_state=sZ(q0) * sZ(q1), out_operator=sY(q0) * sY(q1))
    sz = ExperimentSetting(in_state=sZ(q0) * sZ(q1), out_operator=sZ(q0) * sZ(q1))

    e = Experiment(settings=[sx, sy, sz], program=p)

    results = qc.experiment(e)

    # XX expectation value for bell state |00> + |11> is 1
    assert np.isclose(results[0].expectation, 1)
    assert np.isclose(results[0].std_err, 0)
    assert results[0].total_counts == 40

    # YY expectation value for bell state |00> + |11> is -1
    assert np.isclose(results[1].expectation, -1)
    assert np.isclose(results[1].std_err, 0)
    assert results[1].total_counts == 40

    # ZZ expectation value for bell state |00> + |11> is 1
    assert np.isclose(results[2].expectation, 1)
    assert np.isclose(results[2].std_err, 0)
    assert results[2].total_counts == 40


def asymmetric_ro_model(qubits: list, p00: float = 0.95, p11: float = 0.90) -> NoiseModel:
    aprobs = np.array([[p00, 1 - p00], [1 - p11, p11]])
    aprobs = {q: aprobs for q in qubits}
    return NoiseModel([], aprobs)


def test_qc_calibration_1q(client: Client):
    # noise model with 95% symmetrized readout fidelity per qubit
    noise_model = asymmetric_ro_model([0], 0.945, 0.955)
    qc = get_qc("1q-qvm", client=client)
    qc.qam.noise_model = noise_model

    # bell state program (doesn't matter)
    p = Program()
    p += RESET()
    p += H(0)
    p += CNOT(0, 1)
    p.wrap_in_numshots_loop(10000)

    # Z experiment
    sz = ExperimentSetting(in_state=sZ(0), out_operator=sZ(0))
    e = Experiment(settings=[sz], program=p)

    results = qc.calibrate(e)

    # Z expectation value should just be 1 - 2 * readout_error
    np.isclose(results[0].expectation, 0.9, atol=0.01)
    assert results[0].total_counts == 20000


def test_qc_calibration_2q(client: Client):
    # noise model with 95% symmetrized readout fidelity per qubit
    noise_model = asymmetric_ro_model([0, 1], 0.945, 0.955)
    qc = get_qc("2q-qvm", client=client)
    qc.qam.noise_model = noise_model

    # bell state program (doesn't matter)
    p = Program()
    p += RESET()
    p += H(0)
    p += CNOT(0, 1)
    p.wrap_in_numshots_loop(10000)

    # ZZ experiment
    sz = ExperimentSetting(in_state=sZ(0) * sZ(1), out_operator=sZ(0) * sZ(1))
    e = Experiment(settings=[sz], program=p)

    results = qc.calibrate(e)

    # ZZ expectation should just be (1 - 2 * readout_error_q0) * (1 - 2 * readout_error_q1)
    np.isclose(results[0].expectation, 0.81, atol=0.01)
    assert results[0].total_counts == 40000


def test_qc_joint_expectation(client: Client, dummy_compiler: DummyCompiler):
    qc = QuantumComputer(name="testy!", qam=QVM(client=client), compiler=dummy_compiler)

    # |01> state program
    p = Program()
    p += RESET()
    p += X(0)
    p.wrap_in_numshots_loop(10)

    # ZZ experiment
    sz = ExperimentSetting(in_state=sZ(0) * sZ(1), out_operator=sZ(0) * sZ(1), additional_expectations=[[0], [1]])
    e = Experiment(settings=[sz], program=p)

    results = qc.experiment(e)

    # ZZ expectation value for state |01> is -1
    assert np.isclose(results[0].expectation, -1)
    assert np.isclose(results[0].std_err, 0)
    assert results[0].total_counts == 40
    # Z0 expectation value for state |01> is -1
    assert np.isclose(results[0].additional_results[0].expectation, -1)
    assert results[0].additional_results[1].total_counts == 40
    # Z1 expectation value for state |01> is 1
    assert np.isclose(results[0].additional_results[1].expectation, 1)
    assert results[0].additional_results[1].total_counts == 40


def test_qc_joint_calibration(client: Client):
    # noise model with 95% symmetrized readout fidelity per qubit
    noise_model = asymmetric_ro_model([0, 1], 0.945, 0.955)
    qc = get_qc("2q-qvm", client=client)
    qc.qam.noise_model = noise_model

    # |01> state program
    p = Program()
    p += RESET()
    p += X(0)
    p.wrap_in_numshots_loop(10000)

    # ZZ experiment
    sz = ExperimentSetting(in_state=sZ(0) * sZ(1), out_operator=sZ(0) * sZ(1), additional_expectations=[[0], [1]])
    e = Experiment(settings=[sz], program=p)

    results = qc.experiment(e)

    # ZZ expectation value for state |01> with 95% RO fid on both qubits is about -0.81
    assert np.isclose(results[0].expectation, -0.81, atol=0.01)
    assert results[0].total_counts == 40000
    # Z0 expectation value for state |01> with 95% RO fid on both qubits is about -0.9
    assert np.isclose(results[0].additional_results[0].expectation, -0.9, atol=0.01)
    assert results[0].additional_results[1].total_counts == 40000
    # Z1 expectation value for state |01> with 95% RO fid on both qubits is about 0.9
    assert np.isclose(results[0].additional_results[1].expectation, 0.9, atol=0.01)
    assert results[0].additional_results[1].total_counts == 40000


def test_qc_expectation_on_qvm(client: Client, dummy_compiler: DummyCompiler):
    # regression test for https://github.com/rigetti/forest-tutorials/issues/2
    qc = QuantumComputer(name="testy!", qam=QVM(client=client), compiler=dummy_compiler)

    p = Program()
    theta = p.declare("theta", "REAL")
    p += RESET()
    p += RY(theta, 0)
    p.wrap_in_numshots_loop(10000)

    sx = ExperimentSetting(in_state=sZ(0), out_operator=sX(0))
    e = Experiment(settings=[sx], program=p)

    thetas = [-np.pi / 2, 0.0, np.pi / 2]
    results = []

    # Verify that multiple calls to qc.experiment with the same experiment backed by a QVM that
    # requires_exectutable does not raise an exception.
    for theta in thetas:
        results.append(qc.experiment(e, memory_map={"theta": [theta]}))

    assert np.isclose(results[0][0].expectation, -1.0, atol=0.01)
    assert np.isclose(results[0][0].std_err, 0)
    assert results[0][0].total_counts == 20000

    # bounds on atol and std_err here are a little loose to try and avoid test flakiness.
    assert np.isclose(results[1][0].expectation, 0.0, atol=0.1)
    assert results[1][0].std_err < 0.01
    assert results[1][0].total_counts == 20000

    assert np.isclose(results[2][0].expectation, 1.0, atol=0.01)
    assert np.isclose(results[2][0].std_err, 0)
    assert results[2][0].total_counts == 20000
